import unittest
from ScopeFoundry import BaseApp
from ScopeFoundry.logged_quantity.array_lq import ArrayLQ


class LQConnectionTestApp(BaseApp):

    name = "LQConnectionTestApp"

    def __init__(self, argv):
        BaseApp.__init__(self, argv)

        lq0 = self.settings.New("lq0", dtype=float, ro=False, initial=5)
        lq1 = self.settings.New("lq1", dtype=float, ro=False, initial=5)
        lq2 = self.settings.New("lq2", dtype=float, ro=False, initial=35)
        lq3 = self.settings.New("lq3", dtype=float, ro=True, initial=35)
        lq4 = self.settings.New("lq4", dtype=float, ro=False, initial=35)
        lq_sum = self.settings.New("lq_sum", dtype=float, ro=True, initial=35)
        lq_sum_bidir = self.settings.New(
            "lq_sum_bidir", dtype=float, ro=False, initial=35
        )

        lq_scale = self.settings.New("lq_scale", dtype=float, ro=False)
        lq_scale2 = self.settings.New("lq_scale2", dtype=float, ro=False)

        lq1.connect_to_lq(lq2)

        lq3.connect_lq_math(lq1, lambda x1: x1 + 1)

        lq4.connect_lq_math(
            lq1,
            func=lambda x: x + 5,
            reverse_func=lambda y: y - 5,
        )

        def sum_lq(*vals):
            return sum(vals)

        lq_sum.connect_lq_math(
            (lq1, lq2, lq3, lq4), func=lambda a, b, c, d: a + b + c + d
        )

        def lq_sum_reverse(new_val, old_vals):
            a, b = old_vals
            return new_val - b, b

        lq_sum_bidir.connect_lq_math(
            (lq0, lq1), func=sum_lq, reverse_func=lq_sum_reverse
        )

        lq_scale.connect_lq_math(
            (lq0,),
            func=lambda x: 10 * x,
            reverse_func=lambda y,: [
                0.1 * y,
            ],
        )

        lq_scale2.connect_lq_scale(lq0, 25.0)

        lq_array = self.settings.New(
            "lq_array", dtype=float, is_array=True, initial=[1.0, 2.0, 3.0, 4.0]
        )
        lq_array_element = self.settings.New("lq_array_element", dtype=float)
        lq_array_element.connect_lq_math((lq_array,), lambda arr: arr[1])

        test_array: ArrayLQ = self.settings.New(
            "test_array",
            dtype=float,
            is_array=True,
            ro=False,
            fmt="%1.2f",
            initial=[[147, 111, 100]],
        )

        array_follower = self.settings.New("array_follower", dtype=float)

        test_array.connect_element_follower_lq(array_follower, index=(0, 1), bidir=True)

        self.ui = self.settings.New_UI()

        # self.console_widget.show()
        self.ui.show()


class LQConnectionTest(unittest.TestCase):

    def setUp(self):
        self.app = LQConnectionTestApp([])

    def test_connect_to_lq(self):
        self.app.settings["lq1"] = 99
        self.assertEqual(self.app.settings["lq2"], 99)

    def test_connect_to_lq_rev(self):
        self.app.settings["lq2"] = 999
        self.assertEqual(self.app.settings["lq1"], 999)

    def test_lq_math(self):
        self.app.settings["lq1"] = 100
        self.assertEqual(self.app.settings["lq4"], 105)

    def test_lq_math_rev(self):
        self.app.settings["lq4"] = 1005
        self.assertEqual(self.app.settings["lq1"], 1000)

    def test_lq_math_multivariable(self):
        self.app.settings["lq0"] = 33
        self.app.settings["lq1"] = 66
        self.assertEqual(self.app.settings["lq_sum_bidir"], 33 + 66)

    def test_lq_math_multivariable_rev(self):
        self.app.settings["lq1"] = ini_val_lq1 = 1
        self.app.settings["lq_sum_bidir"] = 1
        self.assertEqual(self.app.settings["lq0"], 1 - ini_val_lq1)
        self.assertEqual(self.app.settings["lq1"], ini_val_lq1)

    def test_lq_scale(self):
        self.app.settings["lq0"] = 1
        self.assertEqual(self.app.settings["lq_scale2"], 25)

    def test_lq_scale_rev(self):
        self.app.settings["lq_scale2"] = 50
        self.assertEqual(self.app.settings["lq0"], 2)

    def test_lq_math_with_array(self):
        self.app.settings["lq_array"] = [100, 200, 300, 400]
        self.assertEqual(self.app.settings["lq_array_element"], 200)

    def test_array_follower_rev(self):
        self.app.settings["test_array"] = [[147, 33, 100]]
        self.assertEqual(self.app.settings["array_follower"], 33)
        self.app.settings["array_follower"] = 55
        self.assertEqual(self.app.settings["test_array"][0][1], 55)


if __name__ == "__main__":
    # unittest.main()
    app = LQConnectionTestApp([])
    app.exec_()



# intended to make a new setup by running:
# python -m ScopeFoundry.tools.new_setup

import os
import re
import shutil
import tempfile
from pathlib import Path


def convert_import_statement(original_import: str) -> str:
    """
    Convert a ScopeFoundry examples import statement to a local module import statement.

    Parameters:
    original_import (str): The original import statement.

    Returns:
    str: The converted import statement.
    """
    pattern = r"from ScopeFoundry\.examples\.((?:\w+\.)*\w+) imp(\w+)"
    replacement = r"from \1 imp\2"
    new_import = re.sub(pattern, replacement, original_import)
    return new_import


def apply_import_statement_conversion(input_file_path: str) -> None:
    """
    Parameters:
    input_file_path (str): The path to the input Python file.
    """
    with tempfile.NamedTemporaryFile("w", delete=False) as temp_file:
        temp_file_path = temp_file.name
        with open(input_file_path, "r") as infile:
            for line in infile:
                temp_file.write(convert_import_statement(line))

    os.replace(temp_file_path, input_file_path)


def copy_scopefoundry_examples() -> None:

    source_dir = Path(__file__).resolve().parent.parent.parent / "examples"
    dest_dir = Path.cwd()

    if not source_dir.exists():
        print(f"Source directory {source_dir} does not exist.")
        return

    for file_path in source_dir.rglob("*"):
        if file_path.is_file():
            if file_path.suffix not in (".py", ".ini", ".json"):
                continue
            relative_path = file_path.relative_to(source_dir)
            dest_file_path = dest_dir / relative_path

            if dest_file_path.exists():
                print(f"Skipped {dest_file_path}, already exists.")
                continue

            dest_file_path.parent.mkdir(parents=True, exist_ok=True)

            shutil.copy2(file_path, dest_file_path)
            apply_import_statement_conversion(dest_file_path)
            print(f"Copied {file_path} to {dest_file_path} and adjusted imports")


def new_app() -> str:
    copy_scopefoundry_examples()
    print(
        """
test installation by calling
    
    python -m example_slowscan_app

rename/modify *app.py file to your liking.
"""
    )
    return "python -m example_slowscan_app"


if __name__ == "__main__":
    new_app()



from ScopeFoundry import HardwareComponent
from ScopeFoundry.scanning import BaseRaster3DSlowScan


class Example3DSlowScanMeasure(BaseRaster3DSlowScan):

    name = "example_3d_scan"

    def scan_specific_setup(self):
        self.stage: HardwareComponent = self.app.hardware["simulon_xyz_stage"]
        self.detector: HardwareComponent = self.app.hardware["noiser_200"]

        self.settings.New("h_axis", str, initial="x", choices="xyz")
        self.settings.New("v_axis", str, initial="y", choices="xyz")
        self.settings.New("z_axis", str, initial="z", choices="xyz")

    def pre_scan_setup(self):
        if self.settings["save_h5"]:
            self.signal_map = self.h5_meas_group.create_dataset(
                name="signal_map", shape=self.scan_shape, dtype=float
            )

    def setup_figure(self):
        super().setup_figure()
        self.ui.details_groupBox.layout().addWidget(
            self.settings.New_UI(include=("h_axis", "v_axis", "z_axis"))
        )

    def move_position_start(self, h, v, z):
        """gets called when z stack is increased layer"""
        self.stage.settings[f"{self.settings['h_axis']}_position"] = h
        self.stage.settings[f"{self.settings['v_axis']}_position"] = v
        self.stage.settings[f"{self.settings['z_axis']}_position"] = z

    def move_position_slow(self, h, v, dh, dv):
        self.stage.settings[f"{self.settings['h_axis']}_position"] = h
        self.stage.settings[f"{self.settings['v_axis']}_position"] = v

    def move_position_fast(self, h, v, dh, dv):
        self.stage.settings[f"{self.settings['h_axis']}_position"] = h
        self.stage.settings[f"{self.settings['v_axis']}_position"] = v

    def collect_pixel(self, pixel_num, k, j, i):
        signal = self.detector.settings.get_lq("signal").read_from_hardware()
        self.display_image_map[k, j, i] = signal
        if self.settings["save_h5"]:
            self.signal_map[k, j, i] = signal



from qtpy import QtWidgets

from ScopeFoundry.tools.features.new_measurement import new_measure
from ScopeFoundry.tools.page import Page


class NewMeasurement(Page):
    def setup(self) -> None:
        self.name = "new measurement"
        self.settings.New("name", dtype=str, initial="my_measurement_name")
        self.settings.New(
            "import",
            str,
            initial="",
            default_widget_factory=QtWidgets.QTextEdit,
            is_clipboardable=True,
        )

        self.operations.new("create", self.new_measurement)
        self.procedure = (
            "define name and press create",
            "name",
            "create",
            "copy paste to your apps setup function",
            # "run",
            "import",
        )

    def new_measurement(self) -> None:
        info = new_measure(self.settings["name"])
        self.settings["import"] = info["ADD_TO_APP"]
        # self.settings["run"] = "python"



"""
Created on $DATE_PRETTY

@author: $AUTHORS
"""

import pyqtgraph as pg
from qtpy import QtCore, QtWidgets

from ScopeFoundry import Measurement, h5_io


class $READOUT_CLASS_NAME(Measurement):

    name = "$READOUT_NAME"

    def setup(self):

        s = self.settings
        s.New("save_h5", bool, initial=True)

        # data structure of the measurement
        self.data = {"y": [2, 5, 2], "img": [[2, 4], [1, 3]]}

    def setup_figure(self):
        # measurement controlls
        cb_layout = QtWidgets.QHBoxLayout()
        cb_layout.addWidget(self.new_start_stop_button())
        cb_layout.addWidget(
            self.settings.New_UI(
                exclude=("activation", "run_state", "profile", "progress")
            )
        )
        header_widget = QtWidgets.QWidget()
        header_layout = QtWidgets.QVBoxLayout(header_widget)
        header_layout.addLayout(cb_layout)

        # plot
        self.graphics_widget = pg.GraphicsLayoutWidget(border=(100, 100, 100))
        self.plot = self.graphics_widget.addPlot(title=self.name)
        self.plot_lines = {}
        self.plot_lines["y"] = self.plot.plot(pen="g")
        # adding an image item
        # self.img_widget = self.graphics_widget.addPlot()
        # self.img_item = pg.ImageItem()
        # self.img_widget.addItem(self.img_item)

        # hw controls (also in the tree)
        # hw = self.app.hardware["$HW_NAME"]
        # hw_ctr = hw.new_control_widgets()

        # ScopeFoundry assumes .ui is the main widget:
        self.ui = QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical)
        self.ui.addWidget(header_widget)
        self.ui.addWidget(self.graphics_widget)
        # self.ui.addWidget(hw_ctr)

    def update_display(self):
        self.plot_lines["y"].setData(self.data["y"])
        # self.img_item.setImage(self.data["img"])

    def run(self):
        print(self.__class__, "run method not implemented")

        # hw = self.app.hardware["$HW_NAME"]
        # self.data['y'] = hw.dev.read_data()

        if self.settings["save_h5"]:
            self.save_h5_data()

    def save_h5_data(self):
        h5_file = h5_io.h5_base_file(app=self.app, measurement=self)
        h5_meas_group = h5_io.h5_create_measurement_group(self, h5_file)
        for k, v in self.data.items():
            h5_meas_group[k] = v
        h5_file.close()



from pathlib import Path

from ScopeFoundry.h5_analyze_with_ipynb import analyze_with_ipynb
from ScopeFoundry.tools.page import Page

ROOT = Path(__file__).parent.parent.parent
ICONS_PATH = ROOT / "base_app/icons"


class WelcomePage(Page):
    def setup(self):
        self.name = "Welcome"
        self.operations.new(
            " analyze with ipynb",
            analyze_with_ipynb,
            icon_path=ICONS_PATH / "jupyter_logo.png",
        )



import time
from datetime import datetime
from pathlib import Path
from typing import TypedDict

from qtpy.QtWidgets import QLabel, QLineEdit, QCheckBox

from .editor_base_ui import EditorBaseUI
from .base_item import BaseItem

ITEMTYPE = "new_dir"
SAVEDIRITEMTYPE = "save_dir_to_parent"


class NewDirKwargs(TypedDict):
    new_dir_name: str
    include_timestamp: bool


class NewDir(BaseItem):
    item_type = ITEMTYPE

    def visit(self):
        t0 = time.time()
        name = self.kwargs["new_dir_name"]
        include_timestamp = self.kwargs.get("include_timestamp", True)
        if include_timestamp or not name:
            sub_dir = f"{datetime.fromtimestamp(t0):%y%m%d_%H%M%S}_{name}"
        else:
            sub_dir = name
        new_dir = Path(self.app.settings["save_dir"]) / sub_dir
        new_dir.mkdir(exist_ok=True)
        self.app.settings["save_dir"] = new_dir.as_posix()

    def _update_appearance(self, text=None) -> str:
        kwargs: NewDirKwargs = self.kwargs
        if text == None:
            if self.kwargs.get("include_timestamp", True):
                text = f"New dir: [Timestamp]_{kwargs['new_dir_name']}"
            else:
                text = f"New dir: {kwargs['new_dir_name']}"
        self.setText(text)
        return text


class NewDirEditorUI(EditorBaseUI):
    item_type = ITEMTYPE
    description = f"creates sub folder and set as save_dir"

    def setup_ui(self):
        self.new_dir_name_lineEdit = QLineEdit()
        self.include_timestamp_cb = QCheckBox()
        self.include_timestamp_cb.setFixedWidth(22)
        self.include_timestamp_cb.setToolTip("include a date in the folder name")
        self.group_box.layout().addWidget(self.include_timestamp_cb)
        self.group_box.layout().addWidget(self.new_dir_name_lineEdit)

    def get_kwargs(self) -> NewDirKwargs:
        val = self.new_dir_name_lineEdit.text()
        include_timestamp = self.include_timestamp_cb.isChecked()
        return {"new_dir_name": val, "include_timestamp": include_timestamp}

    def set_kwargs(self, new_dir_name, include_timestamp=True):
        self.new_dir_name_lineEdit.setText(new_dir_name)
        self.include_timestamp_cb.setChecked(include_timestamp)
        self.new_dir_name_lineEdit.setFocus()


class SaveDirToParent(BaseItem):
    item_type = SAVEDIRITEMTYPE

    def visit(self):
        cur = Path(self.app.settings["save_dir"])
        self.app.settings["save_dir"] = cur.parent.as_posix()


class SaveDirToParentKwargs(TypedDict):
    info: str


class SaveDirToParentEditorUI(EditorBaseUI):
    item_type = SAVEDIRITEMTYPE
    description = "save_dir to parent, designed use in conjunction with new_dir"

    def setup_ui(self):
        self.spacer = QLabel()
        self.layout.addWidget(self.spacer)

    def get_kwargs(self) -> SaveDirToParentKwargs:
        return {"info": "save_dir jumps to parent"}

    def set_kwargs(self, **kwargs):
        self.spacer.setFocus()



from ScopeFoundry.base_app import BaseMicroscopeApp, BaseApp
from .measurement import Measurement
from .hardware import HardwareComponent
from .logged_quantity import LQCollection, LQRange, LQCollection, LoggedQuantity
from .dynamical_widgets.generic_widget import new_widget, add_to_layout
from .dynamical_widgets.tree_widget import new_tree_widget
from .h5_analyze_with_ipynb import analyze_with_ipynb



import random


class Noiser200Dev:
    """provides an interface between your hardware and python."""

    def __init__(self, port) -> None:
        self.port = port
        # establishing a connection to hardware

    # some awesome value read
    def read_signal(self):
        return random.uniform(0, 1)

    # some awesome value read
    def write_voltage(self, voltage):
        print("wrote a voltage", voltage)



import time
import unittest

import numpy as np
import pyqtgraph as pg

from ScopeFoundry import BaseMicroscopeApp, Measurement


class Measure1(Measurement):

    name = "measure1"

    M = 5

    def setup(self):
        self.settings.New("nested_interrupt", dtype=bool, initial=True)

    def run(self):

        m2 = self.app.measurements["measure2"]

        self.current_data = [0]
        self.data = [None] * self.M

        # start M2
        for j in range(self.M):
            if self.interrupt_measurement_called:
                break
            self.set_progress(100 * (j + 1) / (self.M + 1))
            m2.settings["amplitude"] = j + 1
            success = self.start_nested_measure_and_wait(
                measure=m2,
                nested_interrupt=self.settings["nested_interrupt"],
                polling_func=self.m2_polling,
                polling_time=0.5,
            )
            if not success:
                return
            self.data[j] = m2.data_array

    def m2_polling(self):
        m2 = self.app.measurements["measure2"]
        self.current_data = m2.data_array.copy()

    def setup_figure(self):
        self.ui = self.plot = pg.PlotWidget()

        self.current_plotline = self.plot.plot(pen="r")
        self.plot_lines = []
        for j in range(self.M):
            self.plot_lines.append(self.plot.plot())

    def update_display(self):

        self.current_plotline.setData(self.current_data)

        for j in range(self.M):
            if self.data[j] is None:
                self.plot_lines[j].setData([0])
            else:
                self.plot_lines[j].setData(self.data[j])


class Measure2(Measurement):

    name = "measure2"

    def setup(self):

        self.settings.New("amplitude", dtype=float, initial=1.0)
        self.settings.New("run_crash_immediately", dtype=bool, initial=False)
        self.settings.New("run_crash_middle", dtype=bool, initial=False)
        self.settings.New("pre_run_crash", dtype=bool, initial=False)
        self.settings.New("post_run_crash", dtype=bool, initial=False)

        self.settings.New("sleep_time", float, initial=0.001, si=True, unit="s")
        self.data_array = []

    def setup_figure(self):
        self.ui = self.plot = pg.PlotWidget()

        self.plot_line = self.plot.plot()

    def run(self):

        if self.settings["run_crash_immediately"]:
            raise IOError("run_crash_immediately")

        N = 100

        self.data_array = np.zeros(N)

        for i in range(N):
            # print(self.name, i, 'of', N)
            if self.interrupt_measurement_called:
                print(self.name, "interrupted at", i, "of", N)
                break
            self.set_progress(100.0 * i / N)
            self.data_array[i] = self.settings["amplitude"] * np.sin(2 * np.pi * i / N)
            time.sleep(self.settings["sleep_time"])
            if i > 50 and self.settings["run_crash_middle"]:
                raise IOError("run_crash_middle")

    def update_display(self):
        self.plot_line.setData(self.data_array)

    def pre_run(self):
        print(self.name, "pre_run fun!")

        time.sleep(0.2)
        if self.settings["pre_run_crash"]:
            raise IOError("pre_run_crash")
        time.sleep(0.2)
        print(self.name, "pre_run fun done!")

    def post_run(self):
        print(self.name, "post_run fun!")
        time.sleep(0.2)
        if self.settings["post_run_crash"]:
            raise IOError("post_run_crash")
        time.sleep(0.2)
        print(self.name, "post_run fun done!")


class NestMeasureTestApp(BaseMicroscopeApp):

    name = "nested_measure_test"

    def setup(self):

        self.add_measurement(Measure1(self))
        self.add_measurement(Measure2(self))


class NestMeasureTestAppTest(unittest.TestCase):

    def setUp(self):
        # import sys

        self.app = NestMeasureTestApp([])
        self.m1 = self.app.measurements["measure1"]
        self.m2 = self.app.measurements["measure2"]

    def test_no_crash(self):
        self.m2.settings["run_crash_immediately"] = False
        self.m2.settings["run_crash_middle"] = False
        self.m2.settings["pre_run_crash"] = False
        self.m2.settings["post_run_crash"] = False
        self.m1.settings["nested_interrupt"] = False
        self.m1.settings["activation"] = True

        self.assertEqual(self.m1.settings["run_state"], "run_thread_run")
        self.assertEqual(self.m2.settings["run_state"], "stop_first")

    def test_crash_immediately(self):
        self.m2.settings["run_crash_immediately"] = True
        self.m2.settings["run_crash_middle"] = False
        self.m2.settings["pre_run_crash"] = False
        self.m2.settings["post_run_crash"] = False
        self.m1.settings["nested_interrupt"] = False
        self.m1.settings["activation"] = True
        self.assertEqual(self.m1.settings["run_state"], "run_thread_run")
        self.assertEqual(self.m2.settings["run_state"], "stop_first")

    def test_crash_middle(self):
        self.m2.settings["run_crash_immediately"] = False
        self.m2.settings["run_crash_middle"] = True
        self.m2.settings["pre_run_crash"] = False
        self.m2.settings["post_run_crash"] = False
        self.m1.settings["nested_interrupt"] = False
        self.m1.settings["activation"] = True
        self.assertEqual(self.m1.settings["run_state"], "run_thread_run")
        self.assertEqual(self.m2.settings["run_state"], "stop_first")

    def test_pre_run_crash(self):
        self.m2.settings["run_crash_immediately"] = False
        self.m2.settings["run_crash_middle"] = False
        self.m2.settings["pre_run_crash"] = True
        self.m2.settings["post_run_crash"] = False
        self.m1.settings["nested_interrupt"] = False
        self.m1.settings["activation"] = True
        self.assertEqual(self.m1.settings["run_state"], "run_thread_run")
        self.assertEqual(self.m2.settings["run_state"], "stop_first")

    def test_post_run_crash(self):
        self.m2.settings["run_crash_immediately"] = False
        self.m2.settings["run_crash_middle"] = False
        self.m2.settings["pre_run_crash"] = False
        self.m2.settings["post_run_crash"] = True
        self.m1.settings["nested_interrupt"] = False
        self.m1.settings["activation"] = True
        self.assertEqual(self.m1.settings["run_state"], "run_thread_run")
        self.assertEqual(self.m2.settings["run_state"], "stop_first")


if __name__ == "__main__":
    # unittest.main()
    import sys

    app = NestMeasureTestApp(sys.argv)

    app.exec_()



from qtpy import QtWidgets


from functools import partial


class Tools:
    """a grouping that provides tools for *WidgetManagers"""

    # to support more layout types bind add/remove functions here:
    def __init__(self, layout, include, exclude) -> None:
        if isinstance(layout, QtWidgets.QFormLayout):
            self.add_to_layout = partial(_add_to_form_lay, layout=layout)
            self.remove_from_layout = partial(_rm_from_form_lay, layout=layout)
        elif isinstance(layout, QtWidgets.QBoxLayout):
            self.add_to_layout = partial(_add_to_box_lay, layout=layout)
            self.remove_from_layout = partial(_rm_from_box_lay, layout=layout)
        elif isinstance(layout, QtWidgets.QGridLayout):
            self.add_to_layout = partial(_add_to_grid_lay, layout=layout)
            self.remove_from_layout = partial(_rm_from_grid_lay, layout=layout)
        else:
            raise TypeError(f"{type(layout)} not support for dynamical widgets")

        self.include = include
        self.exclude = exclude


def _add_to_form_lay(
    name: str,
    widget: QtWidgets.QWidget,
    layout: QtWidgets.QFormLayout,
) -> None:
    if name is None:
        layout.addRow(widget)
    else:
        layout.addRow(name, widget)


def _rm_from_form_lay(
    name: str,
    widget: QtWidgets.QWidget,
    layout: QtWidgets.QFormLayout,
) -> None:
    row = layout.labelForField(widget)
    row.deleteLater()
    widget.deleteLater()


def _add_to_box_lay(
    name: str, widget: QtWidgets.QWidget, layout: QtWidgets.QBoxLayout
) -> None:
    layout.addWidget(widget)


def _rm_from_box_lay(
    name: str, widget: QtWidgets.QWidget, layout: QtWidgets.QBoxLayout
) -> None:
    layout.takeAt(layout.indexOf(widget))
    widget.deleteLater()


def _add_to_grid_lay(
    name: str, widget: QtWidgets.QWidget, layout: QtWidgets.QGridLayout
) -> None:
    row = layout.rowCount()

    if name is not None:
        label = QtWidgets.QLabel(name)
        layout.addWidget(label, row, 0)
        layout.addWidget(widget, row, 1)
        widget.setMinimumWidth(100)
    else:
        layout.addWidget(widget, row, 0, 1, 2)


def _rm_from_grid_lay(
    name: str, widget: QtWidgets.QWidget, layout: QtWidgets.QGridLayout
) -> None:

    for row in range(layout.rowCount()):
        item = layout.itemAtPosition(row, 0)
        if item is None:
            continue
        label = item.widget()
        if isinstance(label, QtWidgets.QLabel) and label.text() == name:
            layout.removeWidget(label)
            label.deleteLater()
            break

    layout.removeWidget(widget)
    widget.deleteLater()



import unittest

from qtpy import QtWidgets

from ScopeFoundry import BaseMicroscopeApp


class TestApp(BaseMicroscopeApp):

    name = "quickbar_test"

    def setup_ui(self):
        self.add_quickbar(QtWidgets.QLabel("this is a quickbar"))


class QuickbarTest(unittest.TestCase):

    def test_no_crash(self):
        self.app = TestApp()


if __name__ == "__main__":
    # unittest.main()
    import sys

    app = TestApp(sys.argv)

    app.exec_()



import numpy as np
from qtpy import QtCore, QtWidgets

from ScopeFoundry.helper_funcs import get_logger_from_class

from .logged_quantity import LoggedQuantity
from .lq_circular_network import LQCircularNetwork


class LQRange(LQCircularNetwork):
    """
    LQRange is a collection of logged quantities that describe a
    numpy.linspace array inputs.
    Four (or six) LQ's are defined, min, max, num, step (center, span)
    and are connected by signals/slots that keep the quantities
    in sync.
    LQRange.array is the linspace array and is kept upto date
    with changes to the 4 (or 6) LQ's
    """

    def __init__(
        self,
        min_lq: LoggedQuantity,
        max_lq: LoggedQuantity,
        step_lq: LoggedQuantity,
        num_lq: LoggedQuantity,
        center_lq: LoggedQuantity = None,
        span_lq: LoggedQuantity = None,
        sweep_type_lq: LoggedQuantity = None,
    ):
        QtCore.QObject.__init__(self)
        self.log = get_logger_from_class(self)
        self.min = min_lq
        self.max = max_lq
        self.num = num_lq
        self.step = step_lq
        self.center = center_lq
        self.span = span_lq

        lq_dict = {"min": self.min, "max": self.max, "num": self.num, "step": self.step}

        if self.center == None:
            assert self.span == None, "Invalid initialization of LQRange"
        else:
            lq_dict.update({"center": self.center})
        if self.span == None:
            assert self.center == None, "Invalid initialization of LQRange"
        else:
            lq_dict.update({"span": self.span})

        LQCircularNetwork.__init__(self, lq_dict)

        """
        Note: {step, num} and {min,max,span,center} form each 2 circular subnetworks
        The listener functions update the subnetworks, a connect_lq_math connects 
         {min,max,span,center} to {step, num} unidirectional
        """
        self.num.add_listener(self.on_change_num)
        self.step.add_listener(self.on_change_step)
        if self.center and self.span:
            self.center.add_listener(self.on_change_center_span)
            self.span.add_listener(self.on_change_center_span)
            self.min.add_listener(self.on_change_min_max)
            self.max.add_listener(self.on_change_min_max)

        self.step.connect_lq_math((self.min, self.max, self.num), self.calc_step)

        if sweep_type_lq is not None:
            self.sweep_type_map = {
                "up": self.up_sweep_array,
                "down": self.down_sweep_array,
                "up_down": self.up_down_sweep_array,
                "down_up": self.down_up_sweep_array,
                "zig_zag": self.zig_zag_sweep_array,
                "zag_zig": self.zag_zig_sweep_array,
            }
            self.sweep_type = sweep_type_lq
            self.sweep_type.change_choice_list(self.sweep_type_map.keys())
            lq_dict["sweep_type"] = sweep_type_lq

    def calc_num(self, min_, max_, step):
        """
        enforces num to be a positive integer and adjust step accordingly
        returns num,step
        """
        span = max_ - min_
        if step == 0:
            n = 10
        else:
            n = span / step  # num = n+1
            if n < 0:
                n = -n
            n = np.ceil(n)
        step = span / n
        num = n + 1
        return int(num), step

    def calc_step(self, min_, max_, num):
        """
        excludes num=1 to prevent division by zero,
        returns step
        """
        if num == 1:  # prevent division by zero
            num = 2
        step = (max_ - min_) / (num - 1)
        return step

    def calc_span(self, min_, max_):
        return max_ - min_

    def calc_center(self, min_, max_):
        return (max_ - min_) / 2 + min_

    def calc_min(self, center, span):
        return center - span / 2.0

    def calc_max(self, center, span):
        return center + span / 2.0

    def on_change_step(self):
        step = self.step.val
        num, step = self.calc_num(self.min.val, self.max.val, step)
        self.update_values_synchronously(num=num, step=step)

    def on_change_num(self):
        num = self.num.val
        if num == 1:
            num = 2
        step = self.calc_step(self.min.val, self.max.val, self.num.val)
        self.update_values_synchronously(num=num, step=step)

    def on_change_min_max(self):
        min_ = self.min.val
        max_ = self.max.val
        span = self.calc_span(min_, max_)
        center = self.calc_center(min_, max_)
        self.update_values_synchronously(span=span, center=center)

    def on_change_center_span(self):
        span = self.span.val
        center = self.center.val
        min_ = self.calc_min(center, span)
        max_ = self.calc_max(center, span)
        self.update_values_synchronously(min=min_, max=max_)

    @property
    def array(self):
        return np.linspace(self.min.val, self.max.val, self.num.val)

    def zig_zag_sweep_array(self):
        mid_arg = int(self.num.val / 2)
        ar = self.array
        return np.concatenate([ar[mid_arg:], ar[::-1], ar[0:mid_arg]])

    def zag_zig_sweep_array(self):
        return self.zig_zag_sweep_array()[::-1]

    def up_down_sweep_array(self):
        ar = self.array
        return np.concatenate([ar, ar[::-1]])

    def down_up_sweep_array(self):
        return self.up_down_sweep_array()[::-1]

    def down_sweep_array(self):
        return self.array[::-1]

    def up_sweep_array(self):
        return self.array

    @property
    def sweep_array(self):
        if hasattr(self, "sweep_type"):
            return self.sweep_type_map[self.sweep_type.val]()
        else:
            return self.array

    def add_listener(self, func, argtype=(), **kwargs):
        self.min.add_listener(func, argtype, **kwargs)
        self.max.add_listener(func, argtype, **kwargs)
        self.num.add_listener(func, argtype, **kwargs)

    def New_UI(self):
        ui_widget = QtWidgets.QWidget()
        formLayout = QtWidgets.QFormLayout()
        ui_widget.setLayout(formLayout)
        for lqname, lq in self.lq_dict.items():
            formLayout.addRow(lqname, lq.new_default_widget())
        return ui_widget



import datetime
import inspect
import json
import logging
import subprocess
import time
import warnings
from collections import OrderedDict
from functools import partial
from pathlib import Path
from typing import Any, Dict, List, Protocol

from qtpy import QtCore, QtGui, QtWidgets

from ScopeFoundry import h5_io, ini_io
from ScopeFoundry.dynamical_widgets import new_tree_widget, new_widget
from ScopeFoundry.h5_analyze_with_ipynb import generate_ipynb, generate_loaders_py
from ScopeFoundry.helper_funcs import (
    OrderedAttrDict,
    confirm_on_close,
    find_matches,
    get_scopefoundry_version,
    ignore_on_close,
    load_qt_ui_file,
    open_file,
    sibling_path,
)
from ScopeFoundry.logged_quantity import LoggedQuantity, LQCollection

from .base_app import BaseApp
from .logging_handlers import StatusBarHandler, new_log_file_handler
from .show_io_report_dialog import show_io_report_dialog

# useful for developing - may cause problems:
from .base_microscope_app_mdi import Ui_MainWindow


class MeasurementProtocol(Protocol):
    name: str
    settings: LQCollection
    ui: QtWidgets.QWidget
    subwin: QtWidgets.QMdiSubWindow
    activation: LoggedQuantity

    def setup_figure(): ...
    def setup(): ...
    def show_ui(): ...


class HardwareProtocol(Protocol):
    name: str
    settings: LQCollection
    connected: LoggedQuantity

    def connect(): ...
    def disconnect(): ...


class BaseMicroscopeApp(BaseApp):
    name = "ScopeFoundry"
    """The name of the microscope app, default is ScopeFoundry."""
    mdi = True
    """Multiple Document Interface flag. Tells the app whether to include an MDI widget in the app."""

    def __init__(self, argv: List[str] = [], **kwargs: Any) -> None:
        super().__init__(argv, **kwargs)

        self.settings_icon_path = self.icons_path / "settings_logo.png"
        self.jupyter_logo_path = self.icons_path / "jupyter_logo.png"

        self._setup_log_file_handler()
        self._setup_settings_operations(**kwargs)

        # objects to overwrite and populate with setup function.
        self.hardware: Dict[str, HardwareProtocol] = OrderedAttrDict()
        self.measurements: Dict[str, MeasurementProtocol] = OrderedAttrDict()
        self.logo_path = str(self.icons_path / "scopefoundry_logo2B_1024.png")
        self.quickbar = None  # also with self.add_quickbar
        self.setup()

        self._setup_ui_base()
        self._setup_ui_buttons()
        self._setup_ui_tree_column()
        if self.mdi:
            self._setup_ui_mdi_area()
        self._setup_ui_menu_bar()
        self.setup_ui()  # child may overrite
        self._post_setup_ui_quickaccess()
        self._setup_ui_logo()

    def setup(self) -> None:
        """Override to add Hardware and Measurement Components"""
        pass

    def setup_ui(self) -> None:
        """Optional override to set up ui elements after default ui is built"""
        pass

    def _setup_settings_operations(self, **kwargs: Any) -> None:
        initial_save_path = Path.cwd() / "data"
        if not initial_save_path.is_dir():
            initial_save_path.mkdir()
        self.save_dir = self.settings.new_file(
            "save_dir", is_dir=True, initial=initial_save_path.as_posix()
        )
        self.settings.New("sample", dtype=str, initial="")
        self.settings.New(
            name="data_fname_format",
            dtype=str,
            initial="{timestamp:%y%m%d_%H%M%S}_{measurement.name}.{ext}",
        )
        # Potential new alternative default: '{unique_id_short}_{measurement.name}.{ext}'

        self.settings.new_file(
            name="propose_from_file",
            description="right click on setting widget to see and load value from a file",
            file_filters=["Settings (*.ini *.h5)"],
        ).add_listener(self.propose_settings_values_from_file)

        # self.settings.New('log_dir', dtype='file', is_dir=True, initial=initial_log_dir)

        self.setup_dark_mode_option(dark_mode=kwargs.get("dark_mode", None))
        self.add_operation(
            "analyze with ipynb",
            self.on_analyze_with_ipynb,
            "generates h5_data_loaders.py, overview.ipynb, and tries to launch overview.ipynb (vscode with jupyter extension recommended)",
            self.jupyter_logo_path,
        )

    def _setup_log_file_handler(self) -> None:
        log_path = Path.cwd() / "log"
        if not log_path.is_dir():
            log_path.mkdir()
        _fname = f"{self.name}_log_{datetime.datetime.now():%y%m%d_%H%M%S}.txt"
        self.log_file_handler = new_log_file_handler(log_path / _fname)
        logging.getLogger().addHandler(self.log_file_handler)

    def _setup_ui_base(self) -> None:
        """gets called before setup gets called. Could probabliy be called after but this might"""
        if not hasattr(self, "ui_filename"):
            self.ui_filename = sibling_path(__file__, "base_microscope_app_mdi.ui")

        self.ui = load_qt_ui_file(self.ui_filename)
        self.ui: Ui_MainWindow = self.ui  # useful for developing - may cause problems
        self.ui.mdiArea.setVisible(self.mdi)
        self.ui.quickaccess_scrollArea.setVisible(self.mdi)

        if self.mdi:
            self.ui.col_splitter.setStretchFactor(0, 0)
            self.ui.col_splitter.setStretchFactor(1, 1)

        self._loaded_measure_uis = {}

        self.ui.show()
        self.ui.activateWindow()

        confirm_on_close(
            widget=self.ui,
            title=f"Close {self.name}?",
            message=f"Do you wish to shut down {self.name}",
            func_on_close=self.on_close,
        )
        handler = StatusBarHandler(logging.INFO, self.show_status_bar_msg)
        self.log.addHandler(handler)

        with open(self.this_path / "base_app" / "base_microscope_app.qss", "r") as file:
            self.qtapp.setStyleSheet(file.read())

    def show_status_bar_msg(self, msg: str, timeout: int = 2000) -> None:
        self.ui.statusbar.showMessage(msg, timeout)

    def _setup_ui_buttons(self) -> None:
        # since v 1.6 no longer required if default .ui is used.
        if hasattr(self.ui, "console_pushButton"):
            self.ui.console_pushButton.clicked.connect(self.console_widget.show)
            self.ui.console_pushButton.clicked.connect(
                self.console_widget.activateWindow
            )
        if hasattr(self.ui, "settings_autosave_pushButton"):
            self.ui.settings_autosave_pushButton.clicked.connect(
                self.settings_auto_save_ini
            )
        if hasattr(self.ui, "settings_load_last_pushButton"):
            self.ui.settings_load_last_pushButton.clicked.connect(
                self.settings_load_last
            )
        if hasattr(self.ui, "settings_save_pushButton"):
            self.ui.settings_save_pushButton.clicked.connect(self.settings_save_dialog)
        if hasattr(self.ui, "settings_load_pushButton"):
            self.ui.settings_load_pushButton.clicked.connect(self.settings_load_dialog)

    def _setup_ui_tree_column(self) -> None:

        mm_tree = new_tree_widget(self.measurements.values(), ["Measurements", "Value"])
        hw_tree = new_tree_widget(self.hardware.values(), ["Hardware", "Value"])
        app_widget = new_widget(
            obj=self,
            title="app",
            style="form",
            include=("save_dir", "sample"),
        )
        app_widget.setAcceptDrops(True)
        app_widget.dragEnterEvent = self.on_drag_on_app_widget
        app_widget.dropEvent = self.on_drop_on_app_widget
        app_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Minimum,
            QtWidgets.QSizePolicy.Policy.Maximum,
        )
        ipynb_btn = self.operations.new_button("analyze with ipynb")
        ipynb_btn.setText(" analyze")
        settings_btn = QtWidgets.QPushButton(" settings")
        settings_btn.clicked.connect(self.show_app_settings)
        settings_btn.setIcon(QtGui.QIcon(str(self.settings_icon_path)))
        app_widget_btns_layout = QtWidgets.QHBoxLayout()
        app_widget_btns_layout.addWidget(ipynb_btn)
        app_widget_btns_layout.addWidget(settings_btn)
        app_widget.layout().addLayout(app_widget_btns_layout)
        app_widget.setObjectName("app_settings")
        app_widget.setTitle("to inspect drop a .h5 or .ini, to load also press ctrl")

        splitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical)
        splitter.addWidget(hw_tree)
        splitter.addWidget(mm_tree)
        splitter.addWidget(app_widget)
        self.ui.tree_layout.addWidget(splitter)

    def _setup_ui_mdi_area(self) -> None:
        self.ui.mdiArea.setTabsClosable(False)
        self.ui.mdiArea.setTabsMovable(True)

        self.logging_subwin = self.add_mdi_subwin(self.logging_widget, "Log")
        self.console_subwin = self.add_mdi_subwin(self.console_widget, "Console")

        for name, measure in self.measurements.items():
            self.log.debug(f"setting up figure for measurement {name}")
            ui = self.load_measure_ui(measure)
            if ui is not None:
                subwin = self.add_mdi_subwin(ui, measure.name)
                measure.subwin = subwin

    def _load_all_measure_uis(self) -> None:
        for measure in self.measurements.values():
            self.load_measure_ui(measure)
        self.ui.action_load_all_measure_uis.setVisible(False)

    def _setup_ui_menu_bar(self) -> None:
        self.ui.action_set_data_dir.triggered.connect(self.save_dir.file_browser)
        self.ui.action_load_ini.triggered.connect(self.settings_load_dialog)
        self.ui.action_auto_save_ini.triggered.connect(self.settings_auto_save_ini)
        self.ui.action_save_ini.triggered.connect(self.settings_save_dialog)
        self.ui.action_load_last.triggered.connect(self.settings_load_last)
        if self.mdi:
            self.ui.action_console.triggered.connect(
                partial(self.bring_mdi_subwin_to_front, subwin=self.console_subwin)
            )
            self.ui.action_log_viewer.triggered.connect(
                partial(self.bring_mdi_subwin_to_front, subwin=self.logging_subwin)
            )
        else:
            self.ui.action_console.triggered.connect(self.console_widget.show)
            self.ui.action_log_viewer.triggered.connect(self.logging_widget.show)
        self.ui.action_console.setIcon(self.console_widget.windowIcon())
        self.ui.action_log_viewer.setIcon(self.logging_widget.windowIcon())
        self.ui.action_propose_from_file.triggered.connect(
            self.settings.get_lq("propose_from_file").file_browser
        )

        self.ui.action_show_settings.triggered.connect(self.show_app_settings)
        self.ui.action_show_settings.setIcon(QtGui.QIcon(str(self.settings_icon_path)))
        self.ui.action_analyze_with_ipynb.triggered.connect(
            partial(self.on_analyze_with_ipynb, folder=None)
        )
        self.ui.action_analyze_with_ipynb.setIcon(
            QtGui.QIcon(str(self.jupyter_logo_path))
        )

        self.ui.action_docs.triggered.connect(
            partial(
                self.launch_browser, url="https://www.scopefoundry.org/#documentation"
            )
        )
        self.ui.action_about.triggered.connect(self.on_about)
        # Refer to existing ui object:
        self.menubar = self.ui.menuWindow

        self.ui.menuAdvanced.addAction(
            "new hardware", partial(self.start_tools, "new hardware")
        )
        self.ui.menuAdvanced.addAction(
            "new measurement", partial(self.start_tools, "new measurement")
        )

        if self.mdi:
            self.ui.tab_action.triggered.connect(self.set_tab_mode)
            self.ui.window_action.triggered.connect(self.set_subwindow_mode)
            self.ui.cascade_action.triggered.connect(self.cascade_layout)
            self.ui.tile_action.triggered.connect(self.tile_layout)
            self.ui.action_load_window_positions.triggered.connect(
                self.window_positions_load_dialog
            )
            self.ui.action_save_window_positions.triggered.connect(
                self.window_positions_save_dialog
            )
            self.ui.action_load_all_measure_uis.setVisible(False)
        else:
            self.ui.tab_action.setVisible(False)
            self.ui.window_action.setVisible(False)
            self.ui.cascade_action.setVisible(False)
            self.ui.tile_action.setVisible(False)
            self.ui.action_load_window_positions.setVisible(False)
            self.ui.action_save_window_positions.setVisible(False)
            self.ui.action_load_all_measure_uis.triggered.connect(
                self._load_all_measure_uis
            )

    def _post_setup_ui_quickaccess(self) -> None:
        # check again if quickbar is defined.
        if isinstance(self.quickbar, QtWidgets.QWidget):
            self.ui.quickaccess_scrollArea.setVisible(True)
            layout: QtWidgets.QVBoxLayout = self.ui.quickaccess_layout
            if layout.isEmpty():
                layout.addWidget(self.quickbar)
        else:
            self.ui.quickaccess_scrollArea.setVisible(False)

    def _setup_ui_logo(self) -> None:
        logo_icon = QtGui.QIcon(self.logo_path)
        self.qtapp.setWindowIcon(logo_icon)
        self.ui.setWindowIcon(logo_icon)
        self.ui.setWindowTitle(self.name)

    def show(self) -> None:
        """Tells Qt to show the user interface"""
        # self.ui.exec_()
        self.ui.show()

    def __del__(self) -> None:
        self.ui = None

    def set_subwindow_mode(self) -> None:
        """Switches Multiple Document Interface to Subwindowed viewing mode."""
        self.ui.mdiArea.setViewMode(QtWidgets.QMdiArea.ViewMode.SubWindowView)

    def set_tab_mode(self) -> None:
        """Switches Multiple Document Interface to Tabbed viewing mode."""
        self.ui.mdiArea.setViewMode(QtWidgets.QMdiArea.ViewMode.TabbedView)

    def tile_layout(self) -> None:
        """Tiles subwindows in user interface. Specifically in the Multi Document Interface."""
        self.set_subwindow_mode()
        self.ui.mdiArea.tileSubWindows()

    def cascade_layout(self) -> None:
        """Cascades subwindows in user interface. Specifically in the Multi Document Interface."""
        self.set_subwindow_mode()
        self.ui.mdiArea.cascadeSubWindows()

    def bring_measure_ui_to_front(self, measure: MeasurementProtocol) -> None:
        ui = self._loaded_measure_uis.get(measure.name, None)
        if ui is None:
            # measure also has no subwin
            return
        if self.mdi:
            self.bring_mdi_subwin_to_front(measure.subwin)
        else:
            ui.show()

    def load_measure_ui(self, measure: MeasurementProtocol) -> QtWidgets.QWidget:
        if measure.name in self._loaded_measure_uis:
            return self._loaded_measure_uis[measure.name]

        measure.setup_figure()
        if not hasattr(measure, "ui"):
            self._loaded_measure_uis[measure.name] = None
            return None

        self._loaded_measure_uis[measure.name] = measure.ui
        measure.ui.setWindowTitle(measure.name)
        self.ui.menuWindow.addAction(measure.name, measure.show_ui)
        return measure.ui

    def bring_mdi_subwin_to_front(self, subwin: QtWidgets.QMdiSubWindow) -> None:
        view_mode = self.ui.mdiArea.viewMode()
        if view_mode == QtWidgets.QMdiArea.ViewMode.SubWindowView:
            subwin.showNormal()
            subwin.raise_()
        elif view_mode == QtWidgets.QMdiArea.ViewMode.TabbedView:
            subwin.showMaximized()
            subwin.raise_()

    def add_mdi_subwin(
        self, widget: QtWidgets.QWidget, name: str
    ) -> QtWidgets.QMdiSubWindow:
        mdiArea: QtWidgets.QMdiArea = self.ui.mdiArea
        subwin = mdiArea.addSubWindow(
            widget,
            QtCore.Qt.WindowType.CustomizeWindowHint
            | QtCore.Qt.WindowType.WindowTitleHint
            | QtCore.Qt.WindowType.WindowMinMaxButtonsHint,
        )
        ignore_on_close(subwin)
        subwin.setWindowTitle(name)
        subwin.setWindowIcon(widget.windowIcon())
        subwin.show()

        return subwin

    def add_quickbar(self, widget: QtWidgets.QWidget) -> QtWidgets.QWidget:
        self.ui.quickaccess_scrollArea.setVisible(True)
        self.ui.quickaccess_layout.addWidget(widget)
        self.quickbar = widget
        return self.quickbar

    def on_close(self) -> None:
        self.log.info("on_close")
        # disconnect all hardware objects
        for hw in self.hardware.values():
            self.log.info(f"disconnecting {hw.name}")
            try:
                hw.settings["connected"] = False
                hw.settings.disconnect_all_from_hardware()
            except Exception as err:
                self.log.error(f"tried to disconnect {hw.name}: {err}")

    def on_analyze_with_ipynb(self, folder: str = None) -> Path:
        if folder is None:
            folder = self.settings["save_dir"]
        loaders_fname, dset_names = generate_loaders_py(folder)
        ipynb_path = generate_ipynb(folder)
        print("")
        print("generated", loaders_fname, f"with {len(dset_names)} loader(s)")
        print("")
        print("check", ipynb_path)
        print("")
        if ipynb_path.exists():
            open_file(ipynb_path)
        return ipynb_path

    def add_hardware(self, hw: HardwareProtocol) -> HardwareProtocol:
        """Loads a HardwareComponent object into the app.

        If *hw* is a class, rather an instance, create an instance
        and add it to self.hardware
        """
        assert not hw.name in self.hardware.keys()

        # If *hw* is a class, rather an instance, create an instance
        if inspect.isclass(hw):
            hw = hw(app=self)

        self.hardware.add(hw.name, hw)

        self.add_lq_collection_to_settings_path(hw.settings)

        return hw

    def add_hardware_component(self, hw: HardwareProtocol) -> HardwareProtocol:
        # DEPRECATED use add_hardware()
        return self.add_hardware(hw)

    def add_measurement(self, measure: MeasurementProtocol) -> MeasurementProtocol:
        """Loads a Measurement object into the app.

        If *measure* is a class, rather an instance, create an instance
        and add it to self.measurements

        """
        # If *measure* is a class, rather an instance, create an instance
        if inspect.isclass(measure):
            measure = measure(app=self)

        assert not measure.name in self.measurements.keys()

        self.measurements.add(measure.name, measure)

        self.add_lq_collection_to_settings_path(measure.settings)

        return measure

    def add_measurement_component(
        self, measure: MeasurementProtocol
    ) -> MeasurementProtocol:
        # DEPRECATED, use add_measurement()
        return self.add_measurement(measure)

    def settings_save_h5(self, fname: str) -> None:
        """
        Saves h5 file to a file.

        ==============  =========  =============================================
        **Arguments:**  **Type:**  **Description:**
        fname           str        relative path to the filename of the h5 file.
        ==============  =========  =============================================
        """
        with h5_io.h5_base_file(self, fname) as h5_file:
            for measurement in self.measurements.values():
                h5_io.h5_create_measurement_group(measurement, h5_file)
            self.log.info(f"settings saved to {fname}")

    def settings_save_ini(
        self,
        fname: str,
        save_ro: bool = True,
        save_app: bool = True,
        save_hardware: bool = True,
        save_measurements: bool = True,
    ) -> None:
        """
        ==============  =========  ==============================================
        **Arguments:**  **Type:**  **Description:**
        fname           str        relative path to the filename of the ini file.
        ==============  =========  ==============================================
        """
        exclude_patterns = []
        if not save_app:
            exclude_patterns.append("app")
        if not save_hardware:
            exclude_patterns.append("measurement")
        if not save_measurements:
            exclude_patterns.append("hardware")
        paths = self.get_setting_paths(exclude_patterns, exclude_ro=not save_ro)

        settings = self.read_settings(paths, ini_string_value=True)
        ini_io.save_settings(fname, settings)

        self.propose_settings_values(Path(fname).name, settings)

        self.log.info(f"settings saved to {fname}")

    def settings_load_file(self, fname: Path) -> None:
        fname = Path(fname)
        if fname.suffix == ".ini":
            self.settings_load_ini(fname)
        elif fname.suffix == ".h5":
            self.settings_load_h5(fname)

    def settings_load_ini(
        self, fname: str, ignore_hw_connect: bool = False, show_report: bool = True
    ) -> None:
        """
        ==============  =========  ==============================================
        **Arguments:**  **Type:**  **Description:**
        fname           str        relative path to the filename of the ini file.
        ==============  =========  ==============================================
        """
        settings = ini_io.load_settings(fname)

        if ignore_hw_connect:
            settings = {
                k: v for k, v in settings.items() if not k.endswith("connected")
            }

        report = self.write_settings_safe(settings)
        self._report = report  # _report for test purpose

        if show_report:
            show_io_report_dialog(fname, report, self.settings_load_ini)

        self.propose_settings_values(Path(fname).name, settings)
        self.log.info(f"settings loaded from {fname}")

    def settings_load_h5(
        self, fname: str, ignore_hw_connect: bool = False, show_report: bool = True
    ) -> None:
        """
        Loads h5 settings given a filename.

        ==============  =========  ====================================================================================
        **Arguments:**  **Type:**  **Description:**
        fname           str        relative path to the filename of the h5 file.
        ==============  =========  ====================================================================================
        """
        settings = h5_io.load_settings(fname)
        if ignore_hw_connect:
            settings = {
                k: v for k, v in settings.items() if not k.endswith("connected")
            }

        report = self.write_settings_safe(settings)
        self._report = report  # _report for test purpose

        if show_report:
            show_io_report_dialog(fname, report, self.settings_load_h5)

        self.propose_settings_values(Path(fname).name, settings)
        self.log.info(f"settings loaded from {fname}")

    def settings_auto_save_ini(self) -> None:
        """
        Saves the ini file to app/save_dir directory with a time stamp in the filename.
        """
        fname = (
            Path(self.settings["save_dir"])
            / f"{datetime.datetime.now():%y%m%d_%H%M%S}_settings.ini"
        )
        self.settings_save_ini(fname)

    def settings_load_last(self) -> None:
        """
        Loads last saved ini file.
        """
        fnames = list(Path.cwd().glob("*_settings.ini"))
        fnames += list(Path(self.settings["save_dir"]).glob("*_settings.ini"))
        self.settings_load_ini(sorted(fnames)[-1])

    def settings_save_dialog(self) -> None:
        """Opens a save as ini dialogue in the app user interface."""
        fname, selectedFilter = QtWidgets.QFileDialog.getSaveFileName(
            self.ui, "Save Settings file", "", "Settings File (*.ini)"
        )
        if fname:
            self.settings_save_ini(fname)

    def settings_load_dialog(self) -> None:
        """Opens a load ini dialogue in the app user interface"""
        fname, selectedFilter = QtWidgets.QFileDialog.getOpenFileName(
            self.ui, "Open Settings file", "", "Settings File (*.ini *.h5)"
        )
        self.settings_load_file(fname)

    def get_lq(self, path: str) -> LoggedQuantity:
        """
        returns the LoggedQuantity defined by a path string of the form 'section/[component/]setting'
        where section are "mm", "hw" or "app"
        """
        parts = path.split("/")
        section = parts[0]
        if section in ("HW", "hardware"):
            path = f"hw/{parts[1]}/{parts[2]}"
        elif section in ("measurement", "measure", "measurements"):
            path = f"mm/{parts[1]}/{parts[2]}"
        if not path in self._setting_paths:
            print(f"WARNING: {'/'.join(parts)} does not exist")
        return self._setting_paths.get(path, None)

    def read_setting(
        self, path: str, read_from_hardware: bool = True, ini_string_value: bool = False
    ) -> Any:
        lq = self.get_lq(path)
        if read_from_hardware and lq.has_hardware_read:
            lq.read_from_hardware()
        if ini_string_value:
            return lq.ini_string_value()
        return lq.val

    def get_setting_paths(
        self,
        filter_has_hardware_read: bool = False,
        filter_has_hardware_write: bool = False,
        exclude_patterns: List[str] = None,
        exclude_ro: bool = False,
    ) -> List[str]:
        if filter_has_hardware_read and filter_has_hardware_write:
            paths = (
                path
                for path, lq in self._setting_paths.items()
                if lq.has_hardware_read() or lq.has_hardware_write()
            )
        elif filter_has_hardware_read:
            paths = (
                path
                for path, lq in self._setting_paths.items()
                if lq.has_hardware_read()
            )
        elif filter_has_hardware_write:
            paths = (
                path
                for path, lq in self._setting_paths.items()
                if lq.has_hardware_write()
            )
        else:
            paths = self._setting_paths.keys()

        exclude_paths = []
        if exclude_ro:
            exclude_paths += [path for path in paths if self.get_lq(path).ro]
        if exclude_patterns is not None:
            exclude_paths += find_matches(paths, exclude_patterns)

        return [path for path in paths if path not in exclude_paths]

    def read_settings(
        self,
        paths: List[str] = None,
        read_from_hardware: bool = False,
        ini_string_value: bool = False,
    ) -> Dict[str, Any]:
        """returns a dictionary (path, value):
        ================== =========  =============================================================================
        **Arguments:**     **Type:**  **Description:**
        paths              List[str]  paths to setting, if None(default) all paths are used
        read_from_hardware bool       if True, values are read from hardware, else the current value is used
        ================== =========  =============================================================================
        """
        paths = self.get_setting_paths() if paths is None else paths
        return {
            p: self.read_setting(p, read_from_hardware, ini_string_value) for p in paths
        }

    def lq_path(self, path: str) -> LoggedQuantity:
        warnings.warn(
            "App.lq_path deprecated, use App.get_lq instead", DeprecationWarning
        )
        return self.get_lq(path)

    def lq_paths_list(self) -> List[str]:
        warnings.warn(
            "App.lq_paths_list deprecated, use App.get_setting_paths instead",
            DeprecationWarning,
        )
        return self.get_setting_paths()

    @property
    def hardware_components(self) -> Dict[str, HardwareProtocol]:
        warnings.warn(
            "App.hardware_components deprecated, used App.hardware", DeprecationWarning
        )
        return self.hardware

    @property
    def measurement_components(self) -> Dict[str, MeasurementProtocol]:
        warnings.warn(
            "App.measurement_components deprecated, used App.measurements",
            DeprecationWarning,
        )
        return self.measurements

    @property
    def logged_quantities(self) -> Dict[str, LoggedQuantity]:
        warnings.warn(
            "app.logged_quantities deprecated use app.settings", DeprecationWarning
        )
        return self.settings.as_dict()

    def set_window_positions(self, positions: Dict[str, Any]) -> None:
        def restore_win_state(subwin: QtWidgets.QMdiSubWindow, win_state):
            subwin.showNormal()
            if win_state["maximized"]:
                subwin.showMaximized()
            elif win_state["minimized"]:
                subwin.showMinimized()
            else:
                subwin.setGeometry(*win_state["geometry"])

        self.set_subwindow_mode()
        for name, win_state in positions.items():
            if name == "log":
                restore_win_state(self.logging_subwin, win_state)
            elif name == "console":
                restore_win_state(self.console_subwin, win_state)
            elif name == "main":
                restore_win_state(self.ui, win_state)
                self.ui.col_splitter.setSizes(win_state["col_splitter_sizes"])
            elif name.startswith("measurement/"):
                M = self.measurements[name.split("/")[-1]]
                restore_win_state(M.subwin, win_state)

    def get_window_positions(self) -> Dict[str, Any]:
        positions = OrderedDict()

        def qrect_to_tuple(qr):
            return (qr.x(), qr.y(), qr.width(), qr.height())

        def win_state_from_subwin(subwin: QtWidgets.QMdiSubWindow):
            window_state = dict(
                geometry=qrect_to_tuple(subwin.geometry()),
                maximized=subwin.isMaximized(),
                minimized=subwin.isMinimized(),
                fullscreen=subwin.isFullScreen(),
            )
            return window_state

        positions["main"] = win_state_from_subwin(self.ui)
        positions["main"]["col_splitter_sizes"] = self.ui.col_splitter.sizes()

        positions["log"] = win_state_from_subwin(self.logging_subwin)
        positions["console"] = win_state_from_subwin(self.console_subwin)

        for name, M in self.measurements.items():
            if hasattr(M, "ui"):
                positions[f"measurement/{name}"] = win_state_from_subwin(M.subwin)

        return positions

    def save_window_positions_json(self, fname: str) -> None:
        positions = self.get_window_positions()
        with open(fname, "w") as outfile:
            json.dump(positions, outfile, indent=4)

    def load_window_positions_json(self, fname: str) -> None:
        with open(fname, "r") as infile:
            positions = json.load(infile)
        self.set_window_positions(positions)

    def window_positions_load_dialog(self) -> None:
        fname, selectedFilter = QtWidgets.QFileDialog.getOpenFileName(
            self.ui, "Open Window Position file", "", "position File (*.json)"
        )
        self.load_window_positions_json(fname)

    def window_positions_save_dialog(self) -> None:
        """Opens a save as ini dialogue in the app user interface."""
        fname, selectedFilter = QtWidgets.QFileDialog.getSaveFileName(
            self.ui, "Save Window Position file", "", "position File (*.json)"
        )
        if fname:
            self.save_window_positions_json(fname)

    def generate_data_path(
        self, measurement: MeasurementProtocol, ext: str, t: float = None
    ) -> Path:
        if t is None:
            t = time.time()
        f = self.settings["data_fname_format"].format(
            app=self,
            measurement=measurement,
            timestamp=datetime.datetime.fromtimestamp(t),
            ext=ext,
        )
        return Path(self.settings["save_dir"]) / f

    def propose_settings_values_from_file(self, fname: str = None) -> None:
        """
        Adds to proposed_values of LQs.
        """
        if fname is None:
            fname = self.settings["propose_from_file"]
        path = Path(fname)
        if not path.exists() or path.suffix not in (".ini", ".h5"):
            return
        if path.suffix == ".ini":
            settings = ini_io.load_settings(path)
        elif path.suffix == ".h5":
            settings = h5_io.load_settings(path)
        self.propose_settings_values(path.name, settings)

    def launch_browser(self, url: str) -> None:
        QtGui.QDesktopServices.openUrl(QtCore.QUrl(url))

    def on_about(self) -> None:

        with open(self.this_path / "README.md") as f:
            markdown = f.read().replace(
                "ScopeFoundry", f"ScopeFoundry ({get_scopefoundry_version()})", 1
            )

        readme = QtWidgets.QTextEdit()
        readme.setMarkdown(markdown)

        dialog = QtWidgets.QWidget()
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(readme)
        dialog.setLayout(layout)
        dialog.show()

    def on_drag_on_app_widget(self, event: QtGui.QDragEnterEvent) -> None:
        if event.mimeData().hasUrls():
            fname = Path([u.toLocalFile() for u in event.mimeData().urls()][0])
            if fname.suffix in (".ini", ".h5"):
                if event.modifiers() == QtCore.Qt.KeyboardModifier.ControlModifier:
                    event.setDropAction(QtCore.Qt.DropAction.CopyAction)
                    self.show_status_bar_msg("drop to load settings")
                else:
                    event.setDropAction(QtCore.Qt.DropAction.LinkAction)
                    self.show_status_bar_msg(
                        "hold ctrl key to load settings - otherwise values are proposed only."
                    )
                event.accept()
                return
        event.ignore()

    def on_drop_on_app_widget(self, event: QtGui.QDropEvent) -> None:
        fname = Path([u.toLocalFile() for u in event.mimeData().urls()][0])
        if event.modifiers() == QtCore.Qt.KeyboardModifier.ControlModifier:
            self.settings_load_file(fname)
        else:
            self.propose_settings_values_from_file(fname)

    def show_app_settings(self) -> None:
        if not hasattr(self, "_app_settings_widget"):

            app_widget = new_widget(self, "app")
            app_widget.setAcceptDrops(True)
            app_widget.dragEnterEvent = self.on_drag_on_app_widget
            app_widget.dropEvent = self.on_drop_on_app_widget
            app_widget.setObjectName("app_settings")

            save_btn = QtWidgets.QPushButton("save ...")
            save_btn.clicked.connect(self.settings_save_dialog)
            load_btn = QtWidgets.QPushButton("load ...")
            load_btn.clicked.connect(self.settings_load_dialog)
            hlayout = QtWidgets.QHBoxLayout()
            hlayout.addWidget(save_btn)
            hlayout.addWidget(load_btn)

            widget = self._app_settings_widget = QtWidgets.QWidget()
            widget.setSizePolicy(
                QtWidgets.QSizePolicy.Policy.MinimumExpanding,
                QtWidgets.QSizePolicy.Policy.MinimumExpanding,
            )
            widget.setWindowTitle("settings")
            widget.setWindowIcon(QtGui.QIcon(str(self.settings_icon_path)))
            layout = QtWidgets.QVBoxLayout(widget)
            layout.addWidget(app_widget)
            layout.addLayout(hlayout)

        self._app_settings_widget.show()

    def start_tools(self, page: str = "Welcome") -> None:
        subprocess.Popen(
            ["python", "-m", "ScopeFoundry.tools", page],
            close_fds=True,
            start_new_session=True,
        )



import time
import traceback

import numpy as np

from ScopeFoundry import h5_io

from .base_raster_scan import BaseRaster2DScan, BaseRaster3DScan


class BaseRaster2DSlowScan(BaseRaster2DScan):

    name = "base_raster_2Dslowscan"

    def run(self):
        S = self.settings

        # Hardware
        # self.apd_counter_hc = self.app.hardware_components['apd_counter']
        # self.apd_count_rate = self.apd_counter_hc.apd_count_rate
        # self.stage = self.app.hardware_components['dummy_xy_stage']

        # Data File
        # H5

        # Compute data arrays
        self.compute_scan_arrays()

        self.initial_scan_setup_plotting = True

        # Fill display image with nan
        # this allows for pyqtgraph histogram to ignore unfilled data
        # pyqtgraph ImageItem also keeps unfilled data pixels transparent
        self.display_image_map = np.nan * np.zeros(self.scan_shape, dtype=float)

        while not self.interrupt_measurement_called:
            try:
                # h5 data file setup
                self.t0 = time.time()

                if self.settings["save_h5"]:
                    self.h5_file = h5_io.h5_base_file(self.app, measurement=self)
                    self.h5_filename = self.h5_file.filename

                    self.h5_file.attrs["time_id"] = self.t0
                    H = self.h5_meas_group = h5_io.h5_create_measurement_group(
                        self, self.h5_file
                    )

                    # create h5 data arrays
                    H["h_array"] = self.h_array
                    H["v_array"] = self.v_array
                    H["range_extent"] = self.range_extent
                    H["corners"] = self.corners
                    H["imshow_extent"] = self.imshow_extent
                    H["scan_h_positions"] = self.scan_h_positions
                    H["scan_v_positions"] = self.scan_v_positions
                    H["scan_slow_move"] = self.scan_slow_move
                    H["scan_index_array"] = self.scan_index_array

                # start scan
                self.pixel_i = 0
                self.current_scan_index = self.scan_index_array[0]

                self.pixel_time = np.zeros(self.scan_shape, dtype=float)
                if self.settings["save_h5"]:
                    self.pixel_time_h5 = H.create_dataset(
                        name="pixel_time", shape=self.scan_shape, dtype=float
                    )

                self.pre_scan_setup()

                self.move_position_start(
                    self.scan_h_positions[0], self.scan_v_positions[0]
                )

                for self.pixel_i in range(self.Npixels):
                    if self.interrupt_measurement_called:
                        break

                    i = self.pixel_i

                    self.current_scan_index = self.scan_index_array[i]
                    kk, jj, ii = self.current_scan_index

                    h, v = self.scan_h_positions[i], self.scan_v_positions[i]

                    if self.pixel_i == 0:
                        dh = 0
                        dv = 0
                    else:
                        dh = self.scan_h_positions[i] - self.scan_h_positions[i - 1]
                        dv = self.scan_v_positions[i] - self.scan_v_positions[i - 1]

                    if self.scan_slow_move[i]:
                        if self.interrupt_measurement_called:
                            break
                        self.move_position_slow(h, v, dh, dv)
                        if self.settings["save_h5"]:
                            self.h5_file.flush()  # flush data to file every slow move
                        # self.app.qtapp.ProcessEvents()
                        time.sleep(0.01)
                    else:
                        self.move_position_fast(h, v, dh, dv)

                    self.pos = (h, v)
                    # each pixel:
                    # acquire signal and save to data array
                    pixel_t0 = time.time()
                    self.pixel_time[kk, jj, ii] = pixel_t0
                    if self.settings["save_h5"]:
                        self.pixel_time_h5[kk, jj, ii] = pixel_t0
                    self.collect_pixel(self.pixel_i, kk, jj, ii)
                    self.set_progress(100.0 * self.pixel_i / (self.Npixels))
            except Exception as err:
                self.last_err = err
                self.log.error("Failed to Scan {}".format(repr(err)))
                traceback.print_exc()
                # raise(err)
            finally:
                self.post_scan_cleanup()
                if hasattr(self, "h5_file"):
                    print("h5_file", self.h5_file)
                    try:
                        self.h5_file.close()
                    except ValueError as err:
                        self.log.warning("failed to close h5_file: {}".format(err))
                if not self.settings["continuous_scan"]:
                    break
        print(self.name, "done")

    def move_position_start(self, h, v):
        self.stage.settings.x_position.update_value(h)
        self.stage.settings.y_position.update_value(v)

    def move_position_slow(self, h, v, dh, dv):
        self.stage.settings.x_position.update_value(h)
        self.stage.settings.y_position.update_value(v)

    def move_position_fast(self, h, v, dh, dv):
        self.stage.settings.x_position.update_value(h)
        self.stage.settings.y_position.update_value(v)
        # x = self.stage.settings['x_position']
        # y = self.stage.settings['y_position']
        # x = self.stage.settings.x_position.read_from_hardware()
        # y = self.stage.settings.y_position.read_from_hardware()
        # print(x,y)

    def pre_scan_setup(self):
        print(self.name, "pre_scan_setup not implemented")
        # hardware
        # create data arrays
        # update figure

    def collect_pixel(self, pixel_num, k, j, i):
        # collect data
        # store in arrays
        print(self.name, "collect_pixel", pixel_num, k, j, i, "not implemented")

    def post_scan_cleanup(self):
        print(self.name, "post_scan_cleanup not implemented")

    def new_pt_pos(self, x, y):
        self.move_position_start(x, y)


class BaseRaster3DSlowScan(BaseRaster3DScan):

    name = "base_raster_3Dslowscan"

    def run(self):
        S = self.settings

        # Hardware
        # self.apd_counter_hc = self.app.hardware_components['apd_counter']
        # self.apd_count_rate = self.apd_counter_hc.apd_count_rate
        # self.stage = self.app.hardware_components['dummy_xy_stage']

        # Data File
        # H5

        # Compute data arrays
        self.compute_scan_arrays()

        self.initial_scan_setup_plotting = True

        self.display_image_map = np.zeros(self.scan_shape, dtype=float)

        while not self.interrupt_measurement_called:
            try:
                # h5 data file setup
                self.t0 = time.time()

                if self.settings["save_h5"]:
                    self.h5_file = h5_io.h5_base_file(self.app, measurement=self)
                    self.h5_filename = self.h5_file.filename

                    self.h5_file.attrs["time_id"] = self.t0
                    H = self.h5_meas_group = h5_io.h5_create_measurement_group(
                        self, self.h5_file
                    )

                    # create h5 data arrays
                    H["h_array"] = self.h_array
                    H["v_array"] = self.v_array
                    H["z_array"] = self.z_array
                    H["range_extent"] = self.range_extent
                    H["corners"] = self.corners
                    H["imshow_extent"] = self.imshow_extent
                    H["scan_h_positions"] = self.scan_h_positions
                    H["scan_v_positions"] = self.scan_v_positions
                    H["scan_z_positions"] = self.scan_z_positions
                    H["scan_slow_move"] = self.scan_slow_move
                    H["scan_index_array"] = self.scan_index_array

                # start scan
                self.pixel_i = 0
                self.current_scan_index = self.scan_index_array[0]

                self.pixel_time = np.zeros(self.scan_shape, dtype=float)
                if self.settings["save_h5"]:
                    self.pixel_time_h5 = H.create_dataset(
                        name="pixel_time", shape=self.scan_shape, dtype=float
                    )

                self.pre_scan_setup()

                self.move_position_start(
                    self.scan_h_positions[0],
                    self.scan_v_positions[0],
                    self.scan_z_positions[0],
                )

                for self.pixel_i in range(self.Npixels):
                    if self.interrupt_measurement_called:
                        break

                    i = self.pixel_i

                    self.current_scan_index = self.scan_index_array[i]
                    kk, jj, ii = self.current_scan_index

                    h, v, z = (
                        self.scan_h_positions[i],
                        self.scan_v_positions[i],
                        self.scan_z_positions[i],
                    )

                    if self.pixel_i == 0:
                        dh = 0
                        dv = 0
                    else:
                        dh = self.scan_h_positions[i] - self.scan_h_positions[i - 1]
                        dv = self.scan_v_positions[i] - self.scan_v_positions[i - 1]

                    if self.scan_start_move[i]:
                        if self.interrupt_measurement_called:
                            break
                        self.move_position_start(h, v, z)
                        if self.settings["save_h5"]:
                            self.h5_file.flush()  # flush data to file every slow move
                        # self.app.qtapp.ProcessEvents()
                        time.sleep(0.01)
                    elif self.scan_slow_move[i]:
                        if self.interrupt_measurement_called:
                            break
                        self.move_position_slow(h, v, dh, dv)
                        if self.settings["save_h5"]:
                            self.h5_file.flush()  # flush data to file every slow move
                        # self.app.qtapp.ProcessEvents()
                        time.sleep(0.01)
                    else:
                        self.move_position_fast(h, v, dh, dv)

                    self.pos = (h, v)
                    # each pixel:
                    # acquire signal and save to data array
                    pixel_t0 = time.time()
                    self.pixel_time[kk, jj, ii] = pixel_t0
                    if self.settings["save_h5"]:
                        self.pixel_time_h5[kk, jj, ii] = pixel_t0
                    self.collect_pixel(self.pixel_i, kk, jj, ii)
                    self.set_progress(100.0 * self.pixel_i / (self.Npixels))
            except Exception as err:
                self.last_err = err
                self.log.error("Failed to Scan {}".format(repr(err)))
                traceback.print_exc()
                # raise(err)
            finally:
                self.post_scan_cleanup()
                if hasattr(self, "h5_file"):
                    print("h5_file", self.h5_file)
                    try:
                        self.h5_file.close()
                    except ValueError as err:
                        self.log.warning("failed to close h5_file: {}".format(err))
                if not self.settings["continuous_scan"]:
                    break
        print(self.name, "done")

    def move_position_start(self, h, v, z):
        self.stage.settings.x_position.update_value(h)
        self.stage.settings.y_position.update_value(v)
        self.stage.settings.z_position.update_value(z)

    def move_position_slow(self, h, v, dh, dv):
        self.stage.settings.x_position.update_value(h)
        self.stage.settings.y_position.update_value(v)

    def move_position_fast(self, h, v, dh, dv):
        self.stage.settings.x_position.update_value(h)
        self.stage.settings.y_position.update_value(v)
        # x = self.stage.settings['x_position']
        # y = self.stage.settings['y_position']
        # x = self.stage.settings.x_position.read_from_hardware()
        # y = self.stage.settings.y_position.read_from_hardware()
        # print(x,y)

    def pre_scan_setup(self):
        print(self.name, "pre_scan_setup not implemented")
        # hardware
        # create data arrays
        # update figure

    def collect_pixel(self, pixel_num, k, j, i):
        # collect data
        # store in arrays
        print(self.name, "collect_pixel", pixel_num, k, j, i, "not implemented")

    def post_scan_cleanup(self):
        print(self.name, "post_scan_cleanup not implemented")

    def new_pt_pos(self, x, y):
        self.move_position_start(x, y, self.stage.z_position.val)



import logging
import time
from logging import Handler

from qtpy import QtCore

LEVEL_STYLES = {
    "CRITICAL": "color: red;",
    "ERROR": "color: red;",
    "WARNING": "color: orange;",
    "INFO": "color: green;",
    "DEBUG": "color: green;",
    "NOTSET": "",
}


class LoggingQObject(QtCore.QObject):
    # somehow makes handling errors more robust

    new_log_signal = QtCore.Signal((str,))


class HtmlHandler(Handler):
    """sends a signal with the log in html format"""

    def __init__(self, level=logging.NOTSET, buffer_len=500):

        Handler.__init__(self, level=level)
        self.q_object = LoggingQObject()
        self.new_log_signal = self.q_object.new_log_signal

    def emit(self, record: logging.LogRecord):
        log_entry = self.format(record)
        self.new_log_signal.emit(log_entry)

    def format(self, record: logging.LogRecord):
        # timestamp = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
        timestamp = time.strftime("%Y-%m-%dT%H:%M:%S", time.localtime(record.created))
        style = LEVEL_STYLES.get(record.levelname, "")
        return f"""{timestamp} - <span style="{style}">{record.levelname}</span>: <i>{record.name}</i> :<pre>{record.msg}</pre><br>"""


class StatusBarHandler(Handler):

    def __init__(self, level=logging.NOTSET, setter=print):

        Handler.__init__(self, level=level)
        self.setter = setter

    def emit(self, record: logging.LogRecord):
        self.setter(self.format(record))

    def format(self, record: logging.LogRecord):
        return f"🛈 {record.name}: {record.msg}"


def new_log_file_handler(fname):
    handler = logging.FileHandler(str(fname))
    fmt = "%(asctime)s|%(levelname)s|%(name)s|%(message)s"
    datefmt = "%Y-%m-%dT%H:%M:%S"
    formatter = logging.Formatter(fmt, datefmt)
    handler.setFormatter(formatter)
    return handler



import logging
from functools import partial

from qtpy import QtWidgets

from ScopeFoundry import BaseApp
from ScopeFoundry.dynamical_widgets import new_tree_widget
from ScopeFoundry.ndarray_interactive import ArrayLQ_QTableModel

logging.basicConfig(level='DEBUG')


class LQArrayTestApp(BaseApp):

    name = 'LQArrayTestApp'

    def __init__(self, argv=[]):
        super().__init__(argv)

        self.settings.New(
            "test_array",
            dtype=float,
            is_array=True,
            ro=False,
            fmt="%1.2f",
            initial=[[147, 111, 100]],
        )  # ,[208, 8 , 100],[218, 45 , 100],[345, 1500 , 100,],[517, 8 , 100],[772, 300, 100]])

        self.ui = QtWidgets.QScrollArea()
        self.ui.setWidgetResizable(True)
        hlayout = QtWidgets.QHBoxLayout()
        vlayout = QtWidgets.QVBoxLayout()

        self.ui.setLayout(hlayout)

        self.test_array_model = ArrayLQ_QTableModel(
            self.settings.test_array, col_names=["Center", "FWHM", "Amplitude"]
        )
        self.tableView1 = QtWidgets.QTableView()
        self.tableView1.setModel(self.test_array_model)
        vlayout.addWidget(self.tableView1)

        # repeat above
        self.tableView2 = QtWidgets.QTableView()
        self.tableView2.setModel(self.test_array_model)
        vlayout.addWidget(self.tableView2)

        # no header
        self.test_array_model3 = ArrayLQ_QTableModel(self.settings.test_array)
        self.tableView3 = QtWidgets.QTableView()
        # self.tableView3.horizontalHeader().hide()
        self.tableView3.setModel(self.test_array_model3)
        vlayout.addWidget(self.tableView3)

        self.settings.New(
            "test_array4",
            dtype=float,
            is_array=True,
            ro=False,
            fmt="%1.2f",
            initial=[1, 2, 3],
        )
        self.test_array_model4 = ArrayLQ_QTableModel(self.settings.test_array4 )
        self.tableView4 = QtWidgets.QTableView()
        self.tableView4.setModel(self.test_array_model4)
        vlayout.addWidget(self.tableView4)

        self.test_array_model5 = ArrayLQ_QTableModel(self.settings.test_array4 , transpose=True)
        self.tableView5 = QtWidgets.QTableView()
        self.tableView5.setModel(self.test_array_model5)
        vlayout.addWidget(self.tableView5)

        for dtype in [str, bool, int]:
            lq_name = f"test_array_{dtype.__name__}"
            lq = self.settings.New(
                lq_name, dtype=dtype, is_array=True, ro=False, initial=[0, 1, "20", 0]
            )

            tableModel = ArrayLQ_QTableModel(lq, row_names= [dtype.__name__,], fmt=lq.fmt, transpose=True) 
            tableView = QtWidgets.QTableView()
            tableView.setModel(tableModel)
            vlayout.addWidget(tableView)

        #  int does not round
        self.settings.New("test_int", dtype=int, initial=124)

        hlayout.addWidget(new_tree_widget((self,), ["app", ""]))
        hlayout.addLayout(vlayout)
        self.settings_ui = self.settings.New_UI()
        hlayout.addWidget(self.settings_ui)

        self.ui.show()
        # self.console_widget.show()

        self.load_ini_button = QtWidgets.QPushButton("Load INI file")
        vlayout.addWidget(self.load_ini_button)
        self.save_ini_button = QtWidgets.QPushButton("Save INI file")
        vlayout.addWidget(self.save_ini_button)
        self.load_ini_button = QtWidgets.QPushButton("Load saved INI file")
        vlayout.addWidget(self.load_ini_button)

        self.load_ini_button.clicked.connect(
            partial(self.settings_load_ini, fname="lq_array_test.ini")
        )
        self.save_ini_button.clicked.connect(
            partial(self.settings_save_ini, fname="lq_array_test_saved.ini")
        )
        self.load_ini_button.clicked.connect(
            partial(self.settings_load_ini, fname="lq_array_test_saved.ini")
        )


if __name__ == '__main__':
    import sys

    app = LQArrayTestApp(sys.argv)
    app.exec_()



# Modified from Guido van Rossum's xreload.py to work with python 3.12+
# completely ignores 'kind' variable in _extract_code function
# works for purpose of ScopeFoundry, does not work for all cases

"""Alternative to reload().

This works by executing the module in a scratch namespace, and then
patching classes, methods and functions in place.  This avoids the
need to patch instances.  New objects are copied into the target
namespace.

Some of the many limitiations include:

- Global mutable objects other than classes are simply replaced, not patched

- Code using metaclasses is not handled correctly

- Code creating global singletons is not handled correctly

- Functions and methods using decorators (other than classmethod and
  staticmethod) is not handled correctly

- Renamings are not handled correctly

- Dependent modules are not reloaded

- Frozen modules and modules loaded from zip files aren't handled
  correctly

- Classes involving __slots__ are not handled correctly
"""

import importlib
from importlib import reload
import importlib.util
import inspect
import sys


CLASS_STATICS = frozenset(("__dict__", "__doc__", "__module__", "__weakref__"))


def xreload(mod, new_annotations=None):
    """Reload a module in place, updating classes, methods and functions.

    Args:
      mod: a module object

    Returns:
      The (updated) input object itself.
    """
    # pylint: disable=exec-used
    code = _extract_code(mod)
    if code is None:
        # Fall back to built-in reload()
        return reload(mod)
    # Execute the code.  We copy the module dict to a temporary; then
    # clear the module dict; then execute the new code in the module
    # dict; then swap things back and around.  This trick (due to
    # Glyph Lefkowitz) ensures that the (readonly) __globals__
    # attribute of methods and functions is set to the correct dict
    # object.
    modns = mod.__dict__
    tmpns = {
        "__name__": modns["__name__"],
        "__file__": modns["__file__"],
        "__doc__": modns["__doc__"],
    }
    if new_annotations:
        tmpns["__annotations__"] = new_annotations
    exec(code, tmpns)
    # Now we get to the hard part
    _update_scope(modns, tmpns)
    # Update attributes in place
    for name in set(modns) & set(tmpns):
        modns[name] = _update(modns[name], tmpns[name], mod.__name__)
    # Done!
    return mod


def _extract_code(mod):
    modname = mod.__name__
    if modname == "__main__":
        # print(mod.__dict__)
        # filename = mod._dh[0]
        # stream = open(mod._dh[0])
        raise ImportError(
            "reloading module __main__ currently not supported. Move Measrument/Hardware class to a separate file"
        )
    else:
        pkgname = None
        i = modname.rfind(".")
        if i >= 0:
            pkgname, modname = modname[:i], modname[i + 1 :]
        # Compute the search path
        if pkgname:
            # We're not reloading the package, only the module in it
            path = sys.modules[pkgname].__path__  # Search inside the package
        else:
            # Search the top-level module path
            path = None  # Make find_module() uses the default search path
        # Find the module; may raise ImportError
        spec = importlib.util.find_spec(modname, path)
    # kind = importlib.util.PY
    filename = spec.name
    stream = open(spec.origin)

    # Turn it into a code object
    try:
        # Is it Python source code or byte code read from a file?
        # if kind not in (importlib.PY_COMPILED, importlib.PY_SOURCE):
        #     return None
        source = stream.read().strip() + "\n"
        code = compile(source, filename, "exec")
        return code
    finally:
        if stream:
            stream.close()


def _update_scope(oldscope, newscope):
    oldnames = set(oldscope)
    newnames = set(newscope)
    # Add newly introduced names
    for name in newnames - oldnames:
        oldscope[name] = newscope[name]
    # Delete names that are no longer current
    for name in oldnames - newnames:
        if not name.startswith("__"):
            del oldscope[name]


def _update(oldobj, newobj, modname):
    """Update oldobj, if possible in place, with newobj.

    If oldobj is immutable, this simply returns newobj.

    Args:
      oldobj: the object to be updated
      newobj: the object used as the source for the update

    Returns:
      either oldobj, updated in place, or newobj.
    """
    # pylint: disable=too-many-return-statements
    if not isinstance(oldobj, type(newobj)):
        # Cop-out: if the type changed, give up
        return newobj

    if getattr(newobj, "__module__", None) != modname:
        # Do not update objects in-place that have been imported.
        # Just update their references.
        return newobj

    if inspect.isclass(newobj):
        return _update_class(oldobj, newobj, modname)
    if inspect.isfunction(newobj):
        return _update_function(oldobj, newobj, modname)
    # Not something we recognize, just give up
    return newobj


def _closure_changed(oldcl, newcl):
    old = -1 if oldcl is None else len(oldcl)
    new = -1 if newcl is None else len(newcl)
    if old != new:
        return True
    if old > 0 and new > 0:
        for i in range(old):
            same = oldcl[i] == newcl[i]
            if not same:
                return True
    return False


# All of the following functions have the same signature as _update()


def _update_function(oldfunc, newfunc, _):
    """Update a function object."""
    if _closure_changed(oldfunc.__closure__, newfunc.__closure__):
        raise ClosureChanged()
    oldfunc.__code__ = newfunc.__code__
    oldfunc.__defaults__ = newfunc.__defaults__
    _update_scope(oldfunc.__globals__, newfunc.__globals__)
    return oldfunc


def _update_class(oldclass, newclass, modname):
    """Update a class object."""
    olddict = oldclass.__dict__
    newdict = newclass.__dict__
    oldnames = set(olddict)
    newnames = set(newdict)
    for name in newnames - oldnames:
        setattr(oldclass, name, newdict[name])
    for name in oldnames - newnames:
        delattr(oldclass, name)
    for name in oldnames & newnames - CLASS_STATICS:
        try:
            new = getattr(newclass, name)
            if inspect.isfunction(new):
                _update_function(getattr(oldclass, name, None), new, modname)
            else:
                setattr(oldclass, name, new)
        except ClosureChanged:
            # If the closure changed, we need to replace the entire function
            setattr(oldclass, name, new)
    return oldclass


class ClosureChanged(Exception):
    pass



from functools import partial
from pathlib import Path
import re

from qtpy import QtWidgets, QtGui

from .logged_quantity import LoggedQuantity


class FileLQ(LoggedQuantity):
    """
    Specialized str type :class:`LoggedQuantity` that handles
    a filename (or directory) and associated file.
    """

    def __init__(self, name, default_dir=None, is_dir=False, file_filters=(), **kwargs):

        initial = kwargs.pop("initial", None)
        if not initial:
            initial = str(Path.cwd())

        LoggedQuantity.__init__(self, name, dtype=str, initial=initial, **kwargs)

        self.default_dir = default_dir
        self.is_dir = is_dir

        if isinstance(file_filters, str):
            file_filters = [file_filters]
        self.file_filters = file_filters

    def connect_to_browse_widgets(
        self, lineEdit: QtWidgets.QLineEdit, pushButton: QtWidgets.QPushButton
    ):
        self.connect_to_widget(lineEdit)

        lineEdit.setAcceptDrops(True)
        lineEdit.dragEnterEvent = self.on_drag_enter
        lineEdit.dropEvent = partial(self.on_drop, widget=lineEdit)

        pushButton.clicked.connect(self.file_browser)

    def file_browser(self):

        path = Path(self.default_dir) if self.default_dir else Path(self.val)
        if not path.exists():
            path = Path.cwd()

        directory = str(path) if path.is_dir() else str(path.parent)

        if self.is_dir:
            fname = QtWidgets.QFileDialog.getExistingDirectory(directory=directory)
        else:
            fname, _ = QtWidgets.QFileDialog.getOpenFileName(
                parent=None,
                caption=f"Open File for {self.name}",
                dir=directory,
                filter=";;".join([*self.file_filters, "All Files (*)"]),
            )
        self.log.debug(repr(fname))
        if fname:
            self.update_value(fname)

    def new_default_widget(self):
        line_edit = QtWidgets.QLineEdit()
        btn = QtWidgets.QPushButton("...")
        btn.setMaximumWidth(30)
        self.connect_to_browse_widgets(line_edit, btn)
        widget = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout(widget)
        layout.setSpacing(0)
        layout.addWidget(line_edit)
        layout.addWidget(btn)
        return widget

    def on_drag_enter(self, event: QtGui.QDragEnterEvent):
        if event.mimeData().hasUrls():
            fname = Path([u.toLocalFile() for u in event.mimeData().urls()][0])
            valid_suffixes = re.findall(r"\*\.(\w+)", ";;".join([*self.file_filters]))
            if not self.file_filters or fname.suffix[1:] in valid_suffixes:
                event.accept()
                return

        event.ignore()

    def on_drop(self, event: QtGui.QDropEvent, widget: QtWidgets.QLineEdit):
        fname = Path([u.toLocalFile() for u in event.mimeData().urls()][0])
        widget.setText(str(fname))
        widget.editingFinished.emit()



"""
Created on Jul 23, 2014

Modified by Ed Barnard
UI enhancements by Ed Barnard, Alan Buckley
"""

import asyncio
import enum
import logging
import sys
import traceback
from pathlib import Path
from typing import Callable, Dict, List, Any
from warnings import warn

from qtpy import QtCore, QtWidgets, QtGui

from ScopeFoundry import ini_io
from ScopeFoundry.base_app.console_widget import new_console_widget
from ScopeFoundry.base_app.logging_handlers import HtmlHandler
from ScopeFoundry.base_app.logging_widget import LoggingWidget
from ScopeFoundry.dynamical_widgets.tree_widget import SubtreeManager
from ScopeFoundry.helper_funcs import get_logger_from_class
from ScopeFoundry.logged_quantity import LoggedQuantity, LQCollection
from ScopeFoundry.operations import Operations


# See https://riverbankcomputing.com/pipermail/pyqt/2016-March/037136.html
# makes sure that unhandled exceptions in slots don't crash the whole app with PyQt 5.5 and higher
# old version:
## sys.excepthook = traceback.print_exception
# new version to send to logger
def log_unhandled_exception(*exc_info):
    text = "".join(traceback.format_exception(*exc_info))
    logging.critical(f"Unhandled exception: {text}")
    # print(f"Unhandled exception: {text}")


sys.excepthook = log_unhandled_exception


# To fix a bug with jupyter qtconsole for python 3.8
# https://github.com/jupyter/notebook/issues/4613#issuecomment-548992047
if sys.platform == "win32":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


SETTINGS_PATH_TYPE = Dict[str, LoggedQuantity]
class WRITE_RES(enum.Enum):
    SUCCESS = enum.auto()
    MISSING = enum.auto()
    PROTECTED = enum.auto()


class BaseApp(QtCore.QObject):

    name = "ScopeFoundry"

    def __init__(self, argv: List[str] = [], **kwargs: Any) -> None:
        super().__init__()

        self.icons_path = Path(__file__).parent / "icons"

        self._setup_qtapp(argv)

        self.q_object = BaseAppQObject()

        self.setup_logging()

        self.setup_console_widget()

        # containers to be filled
        self._subtree_managers_ = []
        self._widgets_managers_ = []
        self._setting_paths: SETTINGS_PATH_TYPE = {}
        self.operations = Operations()
        self.settings = LQCollection(path="app")
        self.add_lq_collection_to_settings_path(self.settings)

        # self.setup_dark_mode_option(dark_mode=kwargs.get("dark_mode", None))

    def _setup_qtapp(self, argv: List[str]) -> None:
        self.qtapp = QtWidgets.QApplication.instance()
        if not self.qtapp:
            self.qtapp = QtWidgets.QApplication(argv)
        self.qtapp.setApplicationName(self.name)

    @property
    def this_path(self) -> Path:
        """returns the path to ScopeFoundry package"""
        path = Path(__file__)
        return path.parent.parent

    @property
    def this_filename(self) -> str:
        """returns file path of BaseApp"""
        path = Path(__file__)
        return path.name

    def setup_dark_mode_option(self, dark_mode: bool = None) -> None:
        if hasattr(self.qtapp.styleHints(), "setColorScheme"):
            choices = QtCore.Qt.ColorScheme
            if dark_mode is None:
                initial = QtCore.Qt.ColorScheme.Unknown.value
            elif dark_mode:
                initial = QtCore.Qt.ColorScheme.Dark.value
            else:
                initial = QtCore.Qt.ColorScheme.Light.value
            self.settings.New(
                name="dark_mode",
                dtype=int,
                choices=choices,
                initial=initial,
                description=f"<i>{QtCore.Qt.ColorScheme.Unknown.name}</i> let the operating system decide",
            ).add_listener(self.set_color_scheme, argtype=int)
        elif dark_mode:
            warn(
                "dark mode selection only available with Qt6.8+: pip install PyQt6 or pip install --upgrade PyQt6. trying to use qdarkmode",
                RuntimeWarning,
            )
            try:
                import qdarktheme  # pip install pyqtdarktheme

                qdarktheme.setup()
            except Exception as err:
                warn(
                    "trying to use qdarkmode failed. pip install pyqtdarktheme",
                    RuntimeWarning,
                )
            print(f"pyqdarktheme unavailable: {err}")

        else:
            warn(
                "dark mode selection only available with Qt6.8+: pip install PyQt6 or pip install --upgrade PyQt6",
                RuntimeWarning,
            )

    def set_color_scheme(self, choice: int) -> None:
        scheme = {c.value: c for c in QtCore.Qt.ColorScheme}[choice]
        self.qtapp.styleHints().setColorScheme(scheme)

    def exec_(self) -> int:
        return self.qtapp.exec_()

    def setup(self) -> None:
        pass

    def setup_console_widget(self, kernel: Any = None) -> QtWidgets.QWidget:
        try:
            self.console_widget = new_console_widget(self, kernel)
        except Exception as err:
            print(f"failed to setup console widget {err}")
            self.console_widget = QtWidgets.QWidget()
        self.console_widget.setWindowIcon(
            QtGui.QIcon(str(self.icons_path / "console_logo.png"))
        )
        return self.console_widget

    def setup_logging(self) -> None:
        self.log = get_logger_from_class(self)
        self.log.setLevel(logging.INFO)

        logging.basicConfig(level=logging.WARN)
        logging.getLogger("traitlets").setLevel(logging.WARN)
        logging.getLogger("ipykernel.inprocess").setLevel(logging.WARN)
        logging.getLogger("LoggedQuantity").setLevel(logging.WARN)
        logging.getLogger("PyQt5").setLevel(logging.WARN)

        self.logging_widget = LoggingWidget()
        self.logging_widget.setWindowIcon(
            QtGui.QIcon(str(self.icons_path / "log_logo.png"))
        )
        handler = HtmlHandler(level=logging.DEBUG)
        handler.new_log_signal.connect(self.logging_widget.on_new_log)
        logging.getLogger().addHandler(handler)

    def settings_save_ini_ask(self, dir: str = None, save_ro: bool = True) -> str:
        """Opens a Save dialogue asking the user to select a save destination and give the save file a filename. Saves settings to an .ini file."""
        # TODO add default directory, etc
        fname, _ = QtWidgets.QFileDialog.getSaveFileName(
            self.ui, caption="Save Settings", dir="", filter="Settings (*.ini)"
        )
        if fname:
            self.settings_save_ini(fname, save_ro=save_ro)
        return fname

    def settings_load_ini_ask(self, dir: str = None) -> str:
        """Opens a Load dialogue asking the user which .ini file to load into our app settings. Loads settings from an .ini file."""
        # TODO add default directory, etc
        fname, _ = QtWidgets.QFileDialog.getOpenFileName(None, "Settings (*.ini)")
        if fname:
            self.settings_load_ini(fname)
        return fname

    def on_new_subtree(self, subtree: SubtreeManager) -> None: ...

    def on_right_click(self) -> None: ...

    def add_setting_path(self, lq: LoggedQuantity) -> None:
        self._setting_paths[lq.path] = lq

    def remove_setting_path(self, lq: LoggedQuantity) -> None:
        self._setting_paths.pop(lq.path, None)

    def add_lq_collection_to_settings_path(self, settings: LQCollection) -> None:
        settings.q_object.lq_added.connect(self.add_setting_path)
        settings.q_object.lq_removed.connect(self.remove_setting_path)
        for lq in settings.as_dict().values():
            self.add_setting_path(lq)

    def write_setting(self, path: str, value: Any) -> WRITE_RES:
        lq = self.get_lq(path)
        if lq is None:
            return WRITE_RES.MISSING
        lq.update_value(value)
        return WRITE_RES.SUCCESS

    def write_setting_safe(self, path: str, value: Any) -> WRITE_RES:
        lq = self.get_lq(path)
        if lq is None:
            return WRITE_RES.MISSING
        elif lq.protected:
            return WRITE_RES.PROTECTED
        lq.update_value(value)
        return WRITE_RES.SUCCESS

    def get_lq(self, path: str) -> LoggedQuantity:
        """
        returns the LoggedQuantity defined by a path string.
        """
        return self._setting_paths.get(path, None)

    def write_settings_safe(self, settings: Dict[str, Any]) -> Dict[str, WRITE_RES]:
        """
        updates settings based on a dictionary.

        ==============  =========  ====================================================================================
        **Arguments:**  **Type:**  **Description:**
        settings        dict       (path, value) map
        ==============  =========  ====================================================================================
        """
        report = {}
        for path, value in settings.items():
            success = self.write_setting_safe(path, value)
            report[path] = success
        return report

    def settings_save_ini(self, fname: str, save_ro: bool = True) -> None:
        """
        ==============  =========  ==============================================
        **Arguments:**  **Type:**  **Description:**
        fname           str        relative path to the filename of the ini file.
        ==============  =========  ==============================================
        """
        settings = self.read_settings(None, True)
        ini_io.save_settings(fname, settings)

        if not save_ro:
            settings = {k: v for k, v in settings.items() if not self.get_lq(k).ro}

        self.propose_settings_values(Path(fname).name, settings)

        self.log.info(f"ini settings saved to {fname} str")

    def settings_load_ini(self, fname: str) -> Dict[str, Any]:
        """
        ==============  =========  ==============================================
        **Arguments:**  **Type:**  **Description:**
        fname           str        relative path to the filename of the ini file.
        ==============  =========  ==============================================
        """
        settings = ini_io.load_settings(fname)
        self.write_settings_safe(settings)
        self.propose_settings_values(Path(fname).name, settings)
        return settings

    def propose_settings_values(self, name: str, settings: Dict[str, Any]) -> None:
        """
        Adds to proposed_values of LQs.
        proposed_values can be inspected with right click on connected widgets

        ==============  =========  ====================================================================================
        **Arguments:**  **Type:**  **Description:**
        name            str        label of the proposed value
        settings        dict       (path, value) map
        ==============  =========  ====================================================================================
        """
        for path, val in settings.items():
            lq = self.get_lq(path)
            if lq is None:
                continue
            lq.propose_value(name, val)
        self.log.info(f"proposed values {name}")

    def read_settings(
        self, paths: List[str] = None, ini_string_value: bool = False
    ) -> Dict[str, Any]:
        """
        returns a dictionary (path, value):
        ================== =========  =============================================================================
        **Arguments:**     **Type:**  **Description:**
        paths              list[str]  paths to setting, if None(default) all paths are used
        ================== =========  =============================================================================
        """
        paths = self._setting_paths if paths is None else paths
        return {p: self.read_setting(p, ini_string_value) for p in paths}

    def read_setting(self, path: str, ini_string_value: bool = False) -> Any:
        lq = self.get_lq(path)
        if ini_string_value:
            return lq.ini_string_value()
        return lq.val

    def add_operation(
        self,
        name: str,
        op_func: Callable[[], None],
        description: str = "",
        icon_path: str = "",
    ) -> None:
        """
        Create an operation for the App.

        *op_func* is a function that will be called upon operation activation

        operations are typically exposed in the default ScopeFoundry gui via a pushButton

        :type name: str
        :type op_func: QtCore.Slot or Callable without Argument
        :type description: str
        """
        self.operations.new(name, op_func, description, icon_path)

    def remove_operation(self, name: str) -> None:
        self.operations.remove(name)


class BaseAppQObject(QtCore.QObject):
    pass


class TestBaseApp(BaseApp):

    name = "test base app"

    def __init__(self, argv: List[str] = [], **kwargs: Any) -> None:
        super().__init__(argv, **kwargs)

        self.ui = QtWidgets.QWidget()
        self.ui.show()
        self.console_widget.show()
        self.logging_widget.show()
        self.setup_dark_mode_option(kwargs.get("dark_mode", False))


if __name__ == "__main__":
    # app = BaseApp(sys.argv)
    app = TestBaseApp(sys.argv, dark_mode=False)
    sys.exit(app.exec_())



from collections import deque
import json

import numpy as np
from qtpy import QtCore, QtWidgets

from ScopeFoundry.helper_funcs import QLock, get_logger_from_class
from ScopeFoundry.ndarray_interactive import ArrayLQ_QTableModel

from . import LoggedQuantity


class ArrayLQ(LoggedQuantity):
    updated_shape = QtCore.Signal(str)

    def __init__(
        self,
        name,
        dtype=float,
        initial=[],
        fmt="%g",
        si=True,
        ro=False,
        unit=None,
        vmin=-1e12,
        vmax=+1e12,
        choices=None,
        description=None,
        protected=False,
    ):
        QtCore.QObject.__init__(self)

        self.name = name
        self.dtype = dtype
        if self.dtype == str:
            self.val = np.array(initial, dtype=object)
        else:
            self.val = np.array(initial, dtype=dtype)
        self.hardware_read_func = None
        self.hardware_set_func = None
        self.fmt = fmt  # % string formatting string. This is ignored if dtype==str
        if self.dtype == str:
            self.fmt = "%s"
        self.unit = unit
        self.vmin = vmin
        self.vmax = vmax
        self.ro = ro  # Read-Only
        self.choices = choices
        self.description = description
        self.protected = protected

        self.log = get_logger_from_class(self)

        if self.dtype == int:
            self.spinbox_decimals = 0
        else:
            self.spinbox_decimals = 2
        self.reread_from_hardware_after_write = False

        self.oldval = None

        self._in_reread_loop = False  # flag to prevent reread from hardware loops

        self.widget_list = []
        self.listeners = []

        # threading lock
        self.lock = QLock(mode=0)  # mode 0 is non-reentrant lock

        self.is_array = True

        self._tableView = None

        self.prev_vals = deque([], 3)
        self.proposed_values = deque([], 7)

    def same_values(self, v1, v2):
        if v1.shape == v2.shape:
            return np.all(v1 == v2)
            self.log.debug("same_values %s %s" % (v2 - v1, np.all(v1 == v2)))
        else:
            return False

    def change_shape(self, newshape):
        # TODO
        pass

    def string_value(self):
        return json.dumps(self.val.tolist())

    def ini_string_value(self):
        return json.dumps(self.val.tolist())

    def coerce_to_type(self, x):
        if isinstance(x, str):
            x = json.loads(x)
        return np.array(x, dtype=self.dtype)

    def send_display_updates(self, force=False):
        with self.lock:
            self.log.debug(f"{self.name} send_display_updates")
            if force or np.any(self.oldval != self.val):

                str_val = self.string_value()
                self.updated_value[str].emit(str_val)
                self.updated_text_value.emit(str_val)

                # self.updated_value[float].emit(self.val)
                # if self.dtype != float:
                #    self.updated_value[int].emit(self.val)
                # self.updated_value[bool].emit(self.val)
                self.updated_value[()].emit()

                self.oldval = self.val
            else:
                self.log.debug(
                    f"{self.name} send_display_updates skipped (olval!=self.val)={self.oldval != self.val} force={force} oldval={self.oldval} val={self.val}"
                )

    @property
    def array_tableView(self):
        if self._tableView == None:
            self._tableView = self.create_tableView()
            self._tableView.setWindowTitle(self.name)
        return self._tableView

    def create_tableView(self, **kwargs):
        widget = QtWidgets.QTableView()
        # widget.horizontalHeader().hide()
        # widget.verticalHeader().hide()
        model = ArrayLQ_QTableModel(
            self, transpose=(len(self.val.shape) == 1), **kwargs
        )
        widget.setModel(model)
        return widget

    def connect_element_follower_lq(self, lq, index, bidir=True):
        """
        connects an LQ to follow the changes of the array and display a specific element
        """
        # when self (array_lq) is update, update follower lq
        lq.connect_lq_math((self,), func=lambda arr, index=index: arr[index])

        if bidir:
            # when LQ is updated, update element in self
            def on_element_follower_lq():
                old_val = self.value[index]
                new_val = lq.value
                if new_val == old_val:
                    return
                new_arr = self.value.copy()
                new_arr[index] = new_val
                self.update_value(new_arr)

            lq.add_listener(on_element_follower_lq)

    def new_default_widget(self):
        widget = self.create_tableView()
        widget.horizontalHeader().hide()
        widget.verticalHeader().hide()
        return widget



from setuptools import setup

setup(
    name = 'ScopeFoundry',
    
    version = '2.0.1',
    
    description = 'a platform for laboratory equipment control and scientific data analysis',
    long_description =open('README.md', 'r').read(),
    long_description_content_type='text/markdown',
    
    # Author details
    author='Edward S. Barnard',
    author_email='esbarnard@lbl.gov',

    # Choose your license
    license='BSD',
    
    url='https://www.scopefoundry.org/',

    package_dir={'ScopeFoundry': '.'},
    
    packages=['ScopeFoundry',
              'ScopeFoundry.base_app',
              'ScopeFoundry.controlling',
              'ScopeFoundry.data_browser',
                  'ScopeFoundry.data_browser.viewers',
                  'ScopeFoundry.data_browser.plug_ins',
              'ScopeFoundry.dynamical_widgets',
              'ScopeFoundry.examples',
                  'ScopeFoundry.examples.ScopeFoundryHW',
                  'ScopeFoundry.examples.ScopeFoundryHW.bsinc_noiser200',
                  'ScopeFoundry.examples.ScopeFoundryHW.simulon_xyzstage',
              'ScopeFoundry.graphics',
              'ScopeFoundry.graphics.zoomable_map',
              'ScopeFoundry.logged_quantity',
              'ScopeFoundry.scanning',
              'ScopeFoundry.sequencer',
              'ScopeFoundry.sequencer.item_types',              
              'ScopeFoundry.tools',  
                  'ScopeFoundry.tools.features',
                  'ScopeFoundry.tools.templates',
                  'ScopeFoundry.tools.pages',
            ],
    
    #packages=find_packages('.', exclude=['contrib', 'docs', 'tests']),
    #include_package_data=True,  
    
    # include QT ui files and logo and icons
    package_data={
        "": ["*.ui", "*.icns", "*.png", "*.svg", "*.qss", "_LICENSE", "*.md"]
    },
    install_requires=["numpy", "h5py", "qtpy", "uuid7"],
    extras_require={"all": ["qtconsole", "pyqtdarktheme", "qtpy6"]},

    )



from .logged_quantity import LoggedQuantity
from .array_lq import ArrayLQ
from .file_lq import FileLQ
from .lq_3_vector import LQ3Vector
from .lq_range import LQRange
from .collection import LQCollection



from qtpy import QtWidgets

from ScopeFoundry.tools.page import Page
from ScopeFoundry.tools.features.new_hw import ComTypes, new_hw


class NewHardware(Page):
    def setup(self) -> None:
        self.name = "new hardware"
        self.settings.New("company", dtype=str, initial="thorlabs")
        m = self.settings.New(
            "series or model", dtype=str, initial="XX312_Piezo_Controller"
        )
        self.settings.New("authors", dtype=str, initial="Benedikt Ursprung")
        p = self.settings.New("pretty name", dtype=str, initial="")

        p.connect_lq_math((m,), make_pretty_name)

        self.settings.New("communication type", str, choices=ComTypes)
        self.settings.New("overwrite existing", bool, initial=False)

        self.operations.new("create new hardware", self.new_hardware)

        self.settings.New(
            "import",
            dtype=str,
            default_widget_factory=QtWidgets.QTextEdit,
            is_clipboardable=True,
        )
        self.settings.New("run", str, is_cmd=True, is_clipboardable=True)

        self.operations.new("create", self.new_hardware)
        self.procedure = (
            "define name and press create",
            "company",
            "series or model",
            "authors",
            "communication type",
            "pretty name",
            "overwrite existing",
            "create new hardware",
            "run to test & copy paste to your apps setup function",
            "run",
            "import",
        )

    def new_hardware(self) -> None:
        company = self.settings["company"]
        model = self.settings["series or model"]
        overwrite_existing_module = self.settings["overwrite existing"]
        dev_type = self.settings["communication type"]
        authors = self.settings["authors"]
        if authors == "":
            authors = "Mysterion"
            self.settings["authors"] = authors
        pretty_name = self.settings["pretty name"]
        if pretty_name == "":
            pretty_name = make_pretty_name(model)
            self.settings["pretty name"] = pretty_name

        for verboten in (" ", "-"):
            if verboten in model:
                self.settings["series or model"] = self._clean_verboten(model, verboten)
                return
            if verboten in company:
                self.settings["company"] = self._clean_verboten(company, verboten)

                return
            if verboten in pretty_name:
                self.settings["pretty name"] = self._clean_verboten(
                    pretty_name, verboten
                )
                return

        try:
            info = new_hw(
                company,
                model,
                authors,
                pretty_name,
                overwrite_existing_module,
                dev_type,
            )
        except FileExistsError as e:
            print("Failed: " + str(e))
            return

        print(f"""created ScopeFoundryHW/{info['MODULE_NAME']}""")

        self.settings["import"] = info["ADD_TO_APP"]
        self.settings["run"] = (
            f"python -m ScopeFoundryHW.{info['MODULE_NAME']}.{info['TEST_APP_FILE_NAME'].replace('test_app.py', 'test_app')}"
        )

    def _clean_verboten(self, text: str, verboten: str) -> str:
        return "_".join(text.split(verboten))


def make_pretty_name(model: str) -> str:
    s = model.split("_")
    if len(s) >= 2:
        name = "_".join(s[-2:])
    else:
        name = model
    return name.lower()
